package org.processmining.humbertogarciacaballero.models.controller;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;

import org.processmining.humbertogarciacaballero.models.gui.GraphVisualization;
import org.processmining.humbertogarciacaballero.models.gui.MiniGraphView;
import org.processmining.humbertogarciacaballero.models.gui.PetriNetView;
import org.processmining.humbertogarciacaballero.models.gui.components.SoundnessInteractionGUI;
import org.processmining.humbertogarciacaballero.models.gui.controllers.PathControl.PathMode;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.SoundnessInformation;
import org.processmining.humbertogarciacaballero.models.model.commands.AddNodeCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.AddToMarkingCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.ChangeParametersCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.CommandManager;
import org.processmining.humbertogarciacaballero.models.model.commands.CommandManager.CommandManagerState;
import org.processmining.humbertogarciacaballero.models.model.commands.DragCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.HideNodeCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.RedoCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.RemoveFromMarkingCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.RotateAnticlockwiseCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.RotateClockwiseCommand;
import org.processmining.humbertogarciacaballero.models.model.commands.UndoCommand;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayoutOrientation;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LineType;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.emplacer.EmplacingStrategy;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.layerer.LayeringStrategy;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.sequencer.SequencingStrategy;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;
import org.processmining.prabhakardixit.processanalytics.ProcessAnalyticsViz;
import org.processmining.prabhakardixit.processanalytics.ProcessAnalyticsVizViews;

import prefuse.Visualization;
import prefuse.action.layout.Layout;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Tuple;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class PNetVisualizerController extends Observable {

    // Singleton instance
    private static PNetVisualizerController instance = new PNetVisualizerController();

    // Component
    private JComponent component;
    private JPanel statisticsPanel;  // TODO: do sth with the panel

    // Data
    private Graph graph;
    private HierarchicalLayouter layouter;

    // Visualization data
    private GraphVisualization visualization;
    private PetriNetView mainView;
    private MiniGraphView miniView;

    // Information JPanel
    private JPanel infoPanel;
    private SoundnessInteractionGUI soundnessInteractionGUI;

    // Depth of the path
    private int depth;
    private int maxDepth;
    private PathMode pathMode = PathMode.BOTH;

    // Enabled slider
    private boolean enabledSlider;

    // Undo/redo
    private CommandManager commandManager;
    
    private SoundnessInformation soundnessInformation;
    
    private CoverabilityGraph coverabilityGraph;
    
    private List<Node> initialMarking;
    
    //PMD Process analytics 
    ProcessAnalyticsVizViews processAnalytics;		//TODO initialize

	synchronized public static PNetVisualizerController getInstance() {
        return instance;
    }

    private PNetVisualizerController() {
        graph = new Graph(true);
        layouter = new HierarchicalLayouter(GraphVisualization.GRAPH_GROUP);
        commandManager = new CommandManager();
        initialMarking = new ArrayList<>();
        resetDepth();
    }

    public void initController(Graph graph, JFrame component, SoundnessInformation soundnessInformation, CoverabilityGraph cg) {
        this.soundnessInformation = soundnessInformation;
        visualization = new GraphVisualization(graph);
        mainView = new PetriNetView(visualization);
        miniView = new MiniGraphView(visualization);
        soundnessInteractionGUI = new SoundnessInteractionGUI(visualization.getVisualization(), mainView.getDisplay());
        coverabilityGraph = cg;
        this.graph.clear();
        this.graph = graph;
        initialMarking.clear();
        Iterator<Tuple> iter = graph.getNodes().tuples(PNConstants.STARTING_NODE_PREDICATE);
        while (iter.hasNext()) {
        	initialMarking.add(graph.getNode(iter.next().getRow()));
        }
    }

    //PMD process analytics
    public void initController(Graph graph, JFrame component, SoundnessInformation soundnessInformation, CoverabilityGraph cg, ProcessAnalyticsVizViews pa) {
    	this.setProcessAnalytics(pa);
    	initController(graph,component,soundnessInformation,cg);
    }
    //PMD process analytics
    public Graph getData() {
        return graph;
    }

    public LayeringStrategy getLayeringStrategy() {
        return layouter.getLayeringStrategy();
    }

    public void setLayeringStrategy(LayeringStrategy layeringStrategy) {
        layouter.setLayeringStrategy(layeringStrategy);
    }

    public LayoutOrientation getLayoutOrientation() {
        return layouter.getLayoutOrientation();
    }

    public void setLayoutOrientation(LayoutOrientation layoutOrientation) {
        layouter.setLayoutOrientation(layoutOrientation);
    }

    public Layout getLayouter() {
        return layouter;
    }

    public PetriNetView getMainView() {
        return mainView;
    }

    public MiniGraphView getMiniMap() {
        return miniView;
    }

    public void fitViews() {
        visualization.fitLayoutBounds();
    }

    public void executeLayout(boolean animate) {
        visualization.executeLayout(animate);
    }

    public boolean isSimplifierEnabled() {
        return layouter.isSimplifierEnable();
    }

    public void setSimplifierEnable(boolean enable) {
        layouter.setSimplifierEnable(enable);
    }

    public int getSimplifierCrossingThreshold() {
        return layouter.getSimplifierCrossingThreshold();
    }

    public void setSimplifierCrossingThreshold(int crossingThreshold) {
        layouter.setSimplifierCrossingThreshold(crossingThreshold);
    }

    public int getSimplifierTotalCrossingThreshold() {
        return layouter.getSimplifierTotalCrossingThreshold();
    }

    public void setSimplifierTotalCrossingThreshold(int totalCrossingThreshold) {
        layouter.setSimplifierTotalCrossingThreshold(totalCrossingThreshold);
    }

    public int getSimplifierMaxLoops() {
        return layouter.getSimplifierMaxLoops();
    }

    public void setSimplifierMaxLoops(int maxLoops) {
        layouter.setSimplifierMaxLoops(maxLoops);
    }

    public int getIntracellSpacing() {
        return layouter.getIntracellSpacing();
    }

    public void setIntracellSpacing(int intracellSpacing) {
        layouter.setIntracellSpacing(intracellSpacing);
    }

    public int getLayerSpacing() {
        return layouter.getLayerSpacing();
    }

    public void setLayerSpacing(int layerSpacing) {
        layouter.setLayerSpacing(layerSpacing);
    }

    public float getEdgeLandingOffset() {
        return layouter.getEdgeLandingOffset();
    }

    public void setEdgeLandingOffset(float edgeLandingOffset) {
        layouter.setEdgeLandingOffset(edgeLandingOffset);
    }

    public void rotateClockwise() {
        commandManager.executeCommand(new RotateClockwiseCommand(visualization, layouter));
    }

    public void rotateAnticlockwise() {
        commandManager.executeCommand(new RotateAnticlockwiseCommand(visualization, layouter));
    }

    public void setSequencingStrategy(SequencingStrategy sequencingStrategy) {
        layouter.setSequencingStrategy(sequencingStrategy);
    }

    public SequencingStrategy getSequencingStrategy() {
        return layouter.getSequencingStrategy();
    }

    public void setEmplacingStrategy(EmplacingStrategy emplacingStrategy) {
        layouter.setEmplacingStrategy(emplacingStrategy);
    }

    public EmplacingStrategy getEmplacingStrategy() {
        return layouter.getEmplacingStrategy();
    }

    public boolean isSimplifySameLayer() {
        return layouter.isSimplifySameLayer();
    }

    public void setSimplifySameLayer(boolean simplifySameLayer) {
        layouter.setSimplifySameLayer(simplifySameLayer);
    }

    public void setHighlightNeighborsEnabled(boolean enabled) {
        mainView.setHighlightControlEnabled(enabled);
    }

    public void setSpanEnabled(boolean enabled) {
        mainView.setSpanEnabled(enabled);
    }

    public void searchFor(String query) {
        visualization.searchFor(query);
    }

    public Visualization getVisualization() {
        return visualization.getVisualization();
    }

    public LineType getLineType() {
        return layouter.getLineType();
    }

    public void setLineType(LineType lineType) {
        layouter.setLineType(lineType);
        visualization.setLineType(lineType);
    }

    public void setSimplifyShortEdgesEnabled(boolean enabled) {
        layouter.setSimplifyShortEdgesEnabled(enabled);
    }

    public boolean isSimplifyShortEdgesEnabled() {
        return layouter.isSimplifyShortEdgesEnabled();
    }

    public HierarchicalLayoutContext getLayoutContext() {
        return layouter.getContext();
    }

    public boolean isEdgeDirectionEnabled() {
        return layouter.isEdgeDirectionEnabled();
    }

    public void setEdgeDirectionEnabled(boolean enabled) {
        layouter.setEdgeDirectionEnabled(enabled);
        visualization.setEdgeDirectionEnabled(enabled);
    }

    public void setBarDrawingEnabled(boolean enabled) {
        visualization.setBarDecoratorEnabled(enabled);
    }

    public boolean isBarDrawingEnabled() {
        return visualization.isBarDecoratorEnabled();
    }

    public void setLayerSimplifierControlEnabled(boolean enabled) {
        mainView.setLayerSimplifierControlEnabled(enabled);
        visualization.setSimplifierControlEnabled(enabled);
        if (enabled)
            visualization.runSimplifierDecorator();
    }
    
    public void setEdgeSimplifierControlEnabled(boolean enabled) {
        mainView.setEdgeSimplifierControlEnabled(enabled);
        if (enabled)
            visualization.runSimplifierDecorator();
    }

    public void setLoopHighlightEnabled(boolean enabled) {
        visualization.setLoopModeEnabled(enabled);
    }

    public JPanel getInfoPanel() {
        return infoPanel;
    }

    public void setInfoPanel(JPanel infoPanel) {
        this.infoPanel = infoPanel;
    }

    public int getDepth() {
        return depth;
    }

    public void setDepth(int depth) {
        this.depth = depth;
        setChanged();
        notifyObservers(depth);
    }

    public void setSliderEnabled(boolean enabled) {
        enabledSlider = enabled;
        setChanged();
        notifyObservers(enabledSlider);
    }

    public void setMaxDepth(int maxDepth) {
        this.maxDepth = maxDepth;
        setChanged();
        notifyObservers(true);
    }

    public int getMaxDepth() {
        return maxDepth;
    }

    public void setPathMode(PathMode pathMode) {
        this.pathMode = pathMode;
        setChanged();
        notifyObservers(pathMode);
    }

    public PathMode getPathMode() {
        return pathMode;
    }

    public int getTransitionsWidth() {
        return visualization.getTransitionsWidth();
    }

    public void setTransitionsWidth(int width) {
        visualization.setTransitionsWidth(width);
    }

    public int getTransitionsHeight() {
        return visualization.getTransitionsHeight();
    }

    public void setTransitionsHeight(int height) {
        visualization.setTransitionsHeight(height);
    }

    public int getPlacesSize() {
        return visualization.getPlacesSize();
    }

    public void setPlacesSize(int size) {
        visualization.setPlacesSize(size);
    }

    public int getFontSize() {
        return visualization.getFontSize();
    }

    public void setFontSize(int size) {
        visualization.setFontSize(size);
    }

    public int getDegreeFontSize() {
        return visualization.getDegreeFontSize();
    }

    public void setDegreeFontSize(int size) {
        visualization.setDegreeFontSize(size);
    }

    public long getAnimationLength() {
        return visualization.getAnimationLength();
    }

    public void setAnimationLength(long animationLength) {
        visualization.setAnimationLength(animationLength);
    }

    public CommandManagerState undo() {
        return commandManager.executeCommand(new UndoCommand());
    }

    public CommandManagerState redo() {
        return commandManager.executeCommand(new RedoCommand());
    }

    public void dragNode(VisualItem item, Point2D from, Point2D to) {
        commandManager.executeCommand(new DragCommand(visualization, item, from, to));
    }

    public void addCommandManagerObserver(Observer observer) {
        commandManager.addObserver(observer);
    }

    public void setRemoveModeEnabled(boolean enabled) {
        mainView.setRemoveModeEnabled(enabled);
    }

    public void setAddModeEnabled(boolean enabled) {
        mainView.setAddModeEnabled(enabled);
    }

    public void hideNode(NodeItem node) {
        commandManager.executeCommand(new HideNodeCommand(node));
    }

    public void addNewNode(Node node, boolean afterwards) {
        commandManager.executeCommand(new AddNodeCommand(visualization.getVisualization(), graph, node, afterwards));
    }

    public void addNewNode(Node source, Node target) {
        commandManager.executeCommand(new AddNodeCommand(getVisualization(), graph, source, target));
    }

    public void changeLayoutParameters(boolean firstStagesEnabled, LayeringStrategy layeringStrategy,
            LayoutOrientation layoutOrientation, boolean enabledSimplifier, boolean enabledSimplifyAll,
            boolean enabledSimplifyLongEdges, boolean enabledEdgeDirection, boolean enabledBarDrawing,
            boolean enabledSimplifySameLayer, boolean reorderNodes, int crossingThreshold, int totalCrossingThreshold, int maxLoops,
            int intracellSpacing, int layerSpace, float edgeLandingOffset, SequencingStrategy sequencingStrategy,
            EmplacingStrategy emplacingStrategy, LineType lineType, int transitionsWidth, int transitionsHeight, boolean transitionsInOutDegree,
            int placesWidth, int placesHeight, int fontSize, int degreeFontSize, long animationLength, int edgeColor, int strokeColor) {

        commandManager.executeCommand(new ChangeParametersCommand(visualization, firstStagesEnabled, layeringStrategy,
                layoutOrientation, enabledSimplifier, enabledSimplifyAll, enabledSimplifyLongEdges,
                enabledEdgeDirection, enabledBarDrawing, enabledSimplifySameLayer, reorderNodes, crossingThreshold,
                totalCrossingThreshold, maxLoops, intracellSpacing, layerSpace, edgeLandingOffset, sequencingStrategy,
                emplacingStrategy, lineType, transitionsWidth, transitionsHeight, transitionsInOutDegree, placesWidth, placesHeight, fontSize,
                degreeFontSize, animationLength, edgeColor, strokeColor));
    }

    public void resetDepth() {
        setDepth(1);
    }

    public boolean isFirstStagesEnabled() {
        return layouter.isFirstStagesEnabled();
    }

    public void setFirstStagesEnabled(boolean firstStagesEnabled) {
        layouter.setFirstStagesEnabled(firstStagesEnabled);
    }

    public boolean isSimplifyLongEdgesEnabled() {
        return layouter.isSimplifyLongEdgesEnabled();
    }

    public void setSimplifyLongEdgesEnabled(boolean enabled) {
        layouter.setSimplifyLongEdgesEnabled(enabled);
        visualization.setTextBarDecoratorVisibility(enabled);
    }

    public void setNonLiveTransitionsModeEnabled(boolean enabled) {
        soundnessInteractionGUI.setNonLiveMode();
        soundnessInteractionGUI.setEnabledController(enabled);
    }

    public void setUnboundedPlacesModeEnabled(boolean enabled) {
        soundnessInteractionGUI.setUnboundedMode();
        soundnessInteractionGUI.setEnabledController(enabled);
    }

    public void setDeadTransitionsModeEnabled(boolean enabled) {
        soundnessInteractionGUI.setDeadMode();
        soundnessInteractionGUI.setEnabledController(enabled);
    }
    
    public SoundnessInformation getSoundnessInformation() {
        return soundnessInformation;
    }

    public boolean isTransitionsInOutDegreeEnabled() {
        return visualization.isTransitionInOutDegreeEnabled();
    }

    public boolean isReorderNodesEnabled() {
        return layouter.isReorderNodesEnabled();
    }
    
    public void setReorderNodesEnabled(boolean enabled) {
        layouter.setReorderNodesEnabled(enabled);
    }

    public void addToMarking(NodeItem item) {
        commandManager.executeCommand(new AddToMarkingCommand(getVisualization(), item));
    }

    public void removeFromMarking(NodeItem item, boolean removeAll) {
        commandManager.executeCommand(new RemoveFromMarkingCommand(getVisualization(), item, removeAll));
    }
    
    public CoverabilityGraph getCoverabilityGraph() {
    	return coverabilityGraph;
    }

	public void setTooltipControlEnabled(boolean enabled) {
		mainView.setTooltipEnabled(enabled);
	}

	public int getEdgeColor() {
		return visualization.getEdgeDefaultColor();
	}
	
	public void setEdgeColor(int color) {
		visualization.setEdgeDefaultColor(color);
	}

	public int getStrokeColor() {
		return visualization.getShapeBorderColor();
	}
	
	public void setStrokeColor(int color) {
		visualization.setShapeBorderColor(color);
	}
	
	public List<Node> getInitialMarking() {
		return initialMarking;
	}

	public void setStartingNodeDecoratorEnabled(boolean enabled) {
		visualization.setStartingNodeDecoratorEnabled(enabled);
	}

	public void setStatisticsPanel(JPanel panel) {
		statisticsPanel = panel;
	}
	
	// TODO: Use this method
	public JPanel getStatisticsPanel() {
		return statisticsPanel;
	}

	public void setTransitionSelectionEnabled(boolean selected) {
		mainView.setTransitionSelectionEnabled(selected);
	}

	public void setSelectionNodeEnabled(boolean selected) {
		mainView.setSelectNodeControlEnabled(selected);
	}
	

    public ProcessAnalyticsVizViews getProcessAnalytics() {
		return processAnalytics;
	}

	public void setProcessAnalytics(ProcessAnalyticsVizViews processAnalytics) {
		this.processAnalytics = processAnalytics;
	}

	public void setInformationMode(int selected) {
		ProcessAnalyticsViz.mode = selected;
		visualization.setInformationMode(selected);
	}
	
	public void setSelectedActivity(String activity) {
//		ProcessAnalyticsViz.selectedActivity = activity;
		ProcessAnalyticsViz.setSelectedActivity(activity);
		visualization.setSelectedActivity(activity);
		visualization.getVisualization().run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
		visualization.getVisualization().run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
	}

	public void setFrequencyMode(int selectedIndex) {
		ProcessAnalyticsViz.setFrequencyMode(selectedIndex);
		visualization.getVisualization().run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
		visualization.getVisualization().run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);		
	}
}
