package org.processmining.humbertogarciacaballero.models.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.font.TextAttribute;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.KeyStroke;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.gui.decorator.LabelDecoratorLayout;
import org.processmining.humbertogarciacaballero.models.gui.decorator.NodeDegreeLabelDecoratorLayout;
import org.processmining.humbertogarciacaballero.models.gui.render.CustomLabelRenderer;
import org.processmining.humbertogarciacaballero.models.gui.render.CustomShapeRenderer;
import org.processmining.humbertogarciacaballero.models.gui.render.SimplifiedEdgeRenderer;
import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayoutOrientation;
import org.processmining.humbertogarciacaballero.models.util.Utils;

import prefuse.Constants;
import prefuse.Display;
import prefuse.Visualization;
import prefuse.action.Action;
import prefuse.action.ActionList;
import prefuse.action.RepaintAction;
import prefuse.action.assignment.ColorAction;
import prefuse.action.assignment.ShapeAction;
import prefuse.action.assignment.StrokeAction;
import prefuse.action.layout.SpecifiedLayout;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Schema;
import prefuse.render.DefaultRendererFactory;
import prefuse.render.LabelRenderer;
import prefuse.util.ColorLib;
import prefuse.util.FontLib;
import prefuse.util.PrefuseLib;
import prefuse.util.StrokeLib;
import prefuse.util.display.PaintListener;
import prefuse.visual.VisualItem;
import prefuse.visual.expression.InGroupPredicate;

public class LegendGUI extends JDialog {

    private static final long serialVersionUID = 1L;
    private final JPanel contentPanel = new JPanel();
    private Display d;
    private Visualization vis;

    /**
     * Create the dialog.
     */
    public LegendGUI(JFrame frame, Component mainGraphPanel) {
        super(frame);
        setResizable(false);
        setSize(new Dimension(510, 400));
        setTitle("Legend");
        getContentPane().setLayout(new BorderLayout());
        contentPanel.setLayout(new FlowLayout());
        getContentPane().add(contentPanel, BorderLayout.CENTER);
        setDisplay();

        contentPanel.add(d);

        contentPanel.registerKeyboardAction(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                dispose();
            }
        }, "escape", KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), JComponent.WHEN_IN_FOCUSED_WINDOW);

        pack();
        setLocationRelativeTo(mainGraphPanel);
        setModal(true);

        vis.run("color");
        vis.run("layout");
    }

    private void setDisplay() {
        Graph graph = new Graph(true);
        Utils.initSchema(graph);
        graph.getNodes().addColumn("x", double.class);
        graph.getNodes().addColumn("y", double.class);

        Node n1 = graph.addNode();
        n1.setDouble("x", 8);
        n1.setDouble("y", 7);
        Node n2 = graph.addNode();
        n2.set(PNConstants.NODE_TYPE_KEY, NodeType.TRANSITION.toString());
        n2.setInt(PNConstants.NODE_LAYER_KEY, 1);
        n2.setDouble("x", 8);
        n2.setDouble("y", 33);
        graph.addEdge(n1, n2);

        vis = new Visualization();
        vis.add("graph", graph);

        ColorAction fill = new ColorAction("graph.nodes", VisualItem.FILLCOLOR, ColorLib.color(Color.WHITE));
        ColorAction edges = new ColorAction("graph.edges", VisualItem.STROKECOLOR,
                ColorLib.getCategoryPalette(10, 0.9f, 0.2f, 1.f, 1.0f)[1]);
        ColorAction edges2 = new ColorAction("graph.edges", VisualItem.FILLCOLOR,
                ColorLib.getCategoryPalette(10, 0.9f, 0.2f, 1.f, 1.0f)[1]);
        StrokeAction edgeStrokeSize = new StrokeAction("graph.edges",
                StrokeLib.getStroke(PNConstants.STROKE_DEFAULT_WIDTH));
        StrokeAction borderSize = new StrokeAction("graph.nodes",
                StrokeLib.getStroke(PNConstants.STROKE_DEFAULT_WIDTH));
        ColorAction border = new ColorAction("graph.nodes", VisualItem.STROKECOLOR, PNConstants.STROKE_COLOR);
        ShapeAction shape = new ShapeAction("graph.nodes") {
            @Override
            public int getShape(VisualItem item) {
                NodeType type = NodeType.valueOf(item.getString(PNConstants.NODE_TYPE_KEY));
                if (type == NodeType.PLACE) {
                    return Constants.SHAPE_ELLIPSE;
                } else {
                    return PNConstants.SHAPE_TRANSITION;
                }
            }
        };

        ActionList color = new ActionList();
        color.add(fill);
        color.add(edges);
        color.add(edges2);
        color.add(edgeStrokeSize);
        color.add(borderSize);
        color.add(border);
        color.add(shape);
        
        Action orientationAction = new Action() {
            private LayoutOrientation lo = null;
            @Override
            public void run(double frac) {
                if (lo == null) {
                    lo = PNetVisualizerController.getInstance().getLayoutOrientation();
                    PNetVisualizerController.getInstance().setLayoutOrientation(LayoutOrientation.TOP_TO_BOTTOM);   
                } else {
                    PNetVisualizerController.getInstance().setLayoutOrientation(lo);
                    lo = null;
                }
            }
        };
        ActionList layout = new ActionList();
        layout.add(orientationAction);
        layout.add(new SpecifiedLayout("graph", "x", "y"));
        layout.add(new LabelDecoratorLayout("labeldecorator"));
        layout.add(new NodeDegreeLabelDecoratorLayout("indegreedecorator", true));
        layout.add(new NodeDegreeLabelDecoratorLayout("outdegreedecorator", false));
        layout.add(new RepaintAction());
        layout.add(orientationAction);

        // once actions have been created (see above)
        // they are added to the Visualization.
        vis.putAction("color", color);
        vis.putAction("layout", layout);

        CustomShapeRenderer r = new CustomShapeRenderer();
        r.setTransitionsInOutDegreeEnabled(true);
        SimplifiedEdgeRenderer simplifiedRenderer = new SimplifiedEdgeRenderer();
        simplifiedRenderer.setArrowHeadSize(3, 2);
        DefaultRendererFactory factory = new DefaultRendererFactory(r);
        factory.setDefaultEdgeRenderer(simplifiedRenderer);

        factory.add(new InGroupPredicate("labeldecorator"), new CustomLabelRenderer(PNConstants.NODE_LABEL_KEY) {
            public String getText(VisualItem item) {
                return "Label";
            }
        });
        factory.add(new InGroupPredicate("indegreedecorator"), new LabelRenderer() {
            public String getText(VisualItem item) {
                return "X";
            }
        });
        factory.add(new InGroupPredicate("outdegreedecorator"), new LabelRenderer() {
            public String getText(VisualItem item) {
                return "Y";
            }
        });

        final Schema DECORATOR_SCHEMA = PrefuseLib.getVisualItemSchema();
        DECORATOR_SCHEMA.setDefault(VisualItem.INTERACTIVE, false);
        DECORATOR_SCHEMA.setDefault(VisualItem.TEXTCOLOR, PNConstants.FONT_COLOR);
        DECORATOR_SCHEMA.setDefault(VisualItem.FONT, FontLib.getFont("Arial", 2));

        final Schema DEGREE_DECORATOR_SCHEMA = PrefuseLib.getVisualItemSchema();
        DEGREE_DECORATOR_SCHEMA.setDefault(VisualItem.INTERACTIVE, false);
        DEGREE_DECORATOR_SCHEMA.setDefault(VisualItem.TEXTCOLOR, PNConstants.FONT_COLOR);
        DEGREE_DECORATOR_SCHEMA.setDefault(VisualItem.FONT, FontLib.getFont("Arial", 2));

        vis.addDecorators("labeldecorator", "graph.nodes", PNConstants.TRANSITION_NODE_PREDICATE, DECORATOR_SCHEMA);
        vis.addDecorators("indegreedecorator", "graph.nodes", PNConstants.TRANSITION_NODE_PREDICATE,
                DEGREE_DECORATOR_SCHEMA);
        vis.addDecorators("outdegreedecorator", "graph.nodes", PNConstants.TRANSITION_NODE_PREDICATE,
                DEGREE_DECORATOR_SCHEMA);

        vis.setRendererFactory(factory);

        d = new Display(vis);
        d.setBackground(ColorLib.getColor(240, 240, 240));
        d.setHighQuality(true);
        d.zoom(new Point2D.Double(), 10);
        d.setSize(510, 400);
        d.addPaintListener(new PaintListener() {
            @Override
            public void prePaint(Display d, Graphics2D g) {
            }

            @Override
            public void postPaint(Display d, Graphics2D g) {
                g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
                g.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
                g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
                Map<TextAttribute, Integer> fontAttributes = new HashMap<TextAttribute, Integer>();
                fontAttributes.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
                Font boldUnderline = new Font("Arial", Font.PLAIN, 3).deriveFont(fontAttributes);
                g.setFont(boldUnderline);
                AffineTransform original = g.getTransform();
                g.setTransform(d.getTransform());
                // Place info
                g.drawString("Place", 27, 7);

                // Simplified info
                g.drawString("Simplified edge", 21, 17.5f);
                g.setFont(FontLib.getFont("Serif", 2));
                g.drawString("The arrow indicates the direction", 18, 20.5f);
                g.drawString("The color indicates which nodes are connected", 13, 23.5f);

                // Transition info
                g.setFont(boldUnderline);
                g.drawString("Transition", 24, 30);
                g.setFont(FontLib.getFont("Serif", 2));
                g.drawString("X: Incoming degree", 23, 33.5f);
                g.drawString("Y: Outgoing degree", 23, 36.5f);
                g.drawString("Label: Description of node", 20, 39.5f);

                g.setTransform(original);
            }
        });
    }
}
