package org.processmining.humbertogarciacaballero.models.gui;

import org.processmining.humbertogarciacaballero.models.gui.controllers.CustomDragControl;
import org.processmining.humbertogarciacaballero.models.gui.controllers.CustomTooltipControl;
import org.processmining.humbertogarciacaballero.models.gui.controllers.PlaceHoverInformationControl;
import org.processmining.humbertogarciacaballero.models.gui.controllers.RequestFocusControl;
import org.processmining.humbertogarciacaballero.models.gui.controllers.SelectNodeControl;
import org.processmining.humbertogarciacaballero.models.gui.controllers.SelectSpecificTransitionControl;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.action.Action;
import prefuse.activity.Activity;
import prefuse.activity.ActivityAdapter;
import prefuse.controls.ControlAdapter;
import prefuse.controls.PanControl;
import prefuse.controls.WheelZoomControl;
import prefuse.data.tuple.DefaultTupleSet;
import prefuse.visual.VisualItem;

public class PetriNetView extends GraphView {

    private ControlAdapter dragControl;
    private ControlAdapter panControl;
    private WheelZoomControl zoomControl;
    private CustomTooltipControl tooltipControl;
    private ControlAdapter neighborHighlightControl;
    private ControlAdapter layerSimplifierControl;
    private ControlAdapter edgeSimplifierControl;
    private ControlAdapter addNodeControl;
    private ControlAdapter removeNodeControl;
    private ControlAdapter simplifiedEdgeInformationControl;
    private SelectNodeControl selectionControl;

    public PetriNetView(GraphVisualization vis) {
        super(vis, 0, 0);
    }

    protected void setUpDisplay() {
    	d.addControlListener(new RequestFocusControl());
        d.addControlListener(panControl = new PanControl());

        zoomControl = new WheelZoomControl(true, true);
        zoomControl.setMinScale(0.5);
        d.addControlListener(zoomControl);

//        d.addControlListener(neighborHighlightControl = new NodeHighlightControl(vis.getIncomingFocusGroup(),
//                vis.getOutgoingFocusGroup()));
//        neighborHighlightControl.setEnabled(false);

        d.addControlListener(dragControl = new CustomDragControl(PNConstants.VIS_LAYOUT_DECORATORS));

        d.addControlListener(tooltipControl = new CustomTooltipControl(PNConstants.NODE_LABEL_KEY));
        
        d.addControlListener(selectionControl = new SelectNodeControl((VisualItem) vis.getVisualization().getGroup(GraphVisualization.RUBBER_BAND_GROUP).tuples().next()));
        d.addControlListener(new SelectSpecificTransitionControl());

//        d.addControlListener(layerSimplifierControl = new LayerSimplifierControl(
//                GraphVisualization.LAYER_SIMPLIFIER_DECORATOR_NAME));
//        layerSimplifierControl.setEnabled(false);
//
//        d.addControlListener(edgeSimplifierControl = new EdgeSimplifierControl(vis.getVisualization()));
//        edgeSimplifierControl.setEnabled(false);

        // d.addControlListener(
        // new PathControl(vis.getIncomingFocusGroupPathControl(),
        // vis.getOutgoingFocusGroupPathControl()));

        d.addControlListener(new PlaceHoverInformationControl(d));

//        d.addControlListener(addNodeControl = new AddNodeControl());
//        addNodeControl.setEnabled(false);
//
//        d.addControlListener(removeNodeControl = new HideNodeControl(
//                (VisualItem) vis.getVisualization().getGroup(GraphVisualization.RUBBER_BAND_GROUP).tuples().next()));
//        removeNodeControl.setEnabled(false);
        
//        d.addControlListener(simplifiedEdgeInformationControl = new SimplifiedEdgeInformationControl(d));
        
//        d.addControlListener(new MenuController(d));

        d.setHighQuality(true);
    }

//    public void addPanListener(PanListener listener) {
//        d.addPanListener(listener);
//    }

    public void setHighlightControlEnabled(boolean enabled) {
        neighborHighlightControl.setEnabled(enabled);
    }

    public void setSpanEnabled(boolean enabled) {
        panControl.setEnabled(enabled);
        dragControl.setEnabled(enabled);
    }

    public void setLayerSimplifierControlEnabled(boolean enabled) {
        layerSimplifierControl.setEnabled(enabled);
        // edgeSimplifierControl.setEnabled(!enabled);
        if (!enabled) {
            vis.getVisualization().setValue(GraphVisualization.GRAPH_EDGES_GROUP, null, PNConstants.EDGE_SIMPLIFICABLE,
                    true);
        }
    }

    public void setEdgeSimplifierControlEnabled(boolean enabled) {
        edgeSimplifierControl.setEnabled(enabled);
        // layerSimplifierControl.setEnabled(!enabled);
        if (!enabled) {
            vis.getVisualization().setValue(GraphVisualization.GRAPH_EDGES_GROUP, null, PNConstants.EDGE_SIMPLIFICABLE,
                    true);
        }
    }

    public void setAddModeEnabled(boolean enabled) {
        addNodeControl.setEnabled(enabled);
    }

    public void setRemoveModeEnabled(boolean enabled) {
        removeNodeControl.setEnabled(enabled);
        if (!enabled) {
            DefaultTupleSet ts = (DefaultTupleSet) vis.getVisualization()
                    .getGroup(GraphVisualization.SELECTED_NODES);
            if (ts.getTupleCount() > 0) {
                final Action a = vis.getVisualization().getAction(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
                a.addActivityListener(new ActivityAdapter() {
                    public void activityFinished(Activity activity) {
                        vis.getVisualization().run(PNConstants.VIS_REPAINT_ACTIONS_NAME);
                        a.removeActivityListener(this);
                    }
                });
                ts.clear();
            }
        }
    }

    public boolean isLayerSimplifierControlEnabled() {
        return layerSimplifierControl.isEnabled();
    }
    
    public void setTooltipEnabled(boolean enabled) {
    	tooltipControl.setEnabled(enabled);
    	if (!enabled) tooltipControl.resetTooltip();
    }

    public void setSelectNodeControlEnabled(boolean enabled) {
    	selectionControl.setEnabled(enabled);
    	setSpanEnabled(!enabled);
    	
        if (!enabled) {
            DefaultTupleSet ts = (DefaultTupleSet) vis.getVisualization()
                    .getGroup(GraphVisualization.SELECTED_NODES);
            if (ts.getTupleCount() > 0) {
                final Action a = vis.getVisualization().getAction(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
                a.addActivityListener(new ActivityAdapter() {
                    public void activityFinished(Activity activity) {
                        vis.getVisualization().run(PNConstants.VIS_REPAINT_ACTIONS_NAME);
                        a.removeActivityListener(this);
                    }
                });
                ts.clear();
            }
        }
    }
    
	public void setTransitionSelectionEnabled(boolean enabled) {
		selectionControl.setTransitionSelection(enabled);
        if (!enabled) {
            DefaultTupleSet ts = (DefaultTupleSet) vis.getVisualization()
                    .getGroup(GraphVisualization.SELECTED_NODES);
            if (ts.getTupleCount() > 0) {
                final Action a = vis.getVisualization().getAction(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
                a.addActivityListener(new ActivityAdapter() {
                    public void activityFinished(Activity activity) {
                        vis.getVisualization().run(PNConstants.VIS_REPAINT_ACTIONS_NAME);
                        a.removeActivityListener(this);
                    }
                });
                ts.clear();
            }
        }
	}
}
