package org.processmining.humbertogarciacaballero.models.gui.action;

import java.util.Iterator;

import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.action.GroupAction;
import prefuse.data.Tuple;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.data.tuple.TupleSet;
import prefuse.data.util.Sort;
import prefuse.visual.AggregateItem;
import prefuse.visual.AggregateTable;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class AggregateNodesByLayer extends GroupAction {

	protected AggregateTable aggregateTable;

	public AggregateNodesByLayer(String group, AggregateTable aggregateTable) {
		super(group);
		this.aggregateTable = aggregateTable;
	}

	@SuppressWarnings("unchecked")
	@Override
	public void run(double frac) {
		TupleSet ts = getVisualization().getGroup(getGroup());
		int maxLayer = getMaxLayer(ts);

		for (int i = 0; i <= maxLayer; i++) {
			AggregateItem aggregateItem = getAggregate(i);
			aggregateItem.removeAllItems();
			Iterator<Tuple> iter = ts.tuples(ExpressionParser.predicate("[" + PNConstants.NODE_LAYER_KEY + "] = " + i));
			if (iter.hasNext()) {
				while (iter.hasNext()) {
					VisualItem vi = (VisualItem) iter.next();
					if (!aggregateItem.containsItem(vi)) {
						aggregateItem.addItem(vi);
					}
				}
				aggregateItem.setInt(PNConstants.NODE_LAYER_KEY, i);
				aggregateItem.setInt(PNConstants.NODE_MAX_LAYER_KEY, maxLayer);
				aggregateItem.setBoolean(PNConstants.EDGE_LONG_EDGE_KEY, containsLongEdges(ts, i));
			}
		}
	}

	protected boolean containsLongEdges(TupleSet ts, int i) {
		boolean longEdges = false;

		Iterator<Tuple> iter = ts
				.tuples(ExpressionParser.predicate("([" + PNConstants.SIMPLIFIED_BOTTOM_KEY + "] = TRUE AND ["
						+ PNConstants.NODE_LAYER_KEY + "] = " + i + ") OR ([" + PNConstants.SIMPLIFIED_TOP_KEY
						+ "] = TRUE AND [" + PNConstants.NODE_LAYER_KEY + "] = " + (i + 1) + ")"));
		while (iter.hasNext()) {
			NodeItem ni = (NodeItem) iter.next();
			Iterator<EdgeItem> edges = ni.edges();
			while (edges.hasNext()) {
				EdgeItem edge = edges.next();
				NodeItem adjacent = edge.getAdjacentItem(ni);
				if (edge.getBoolean(PNConstants.EDGE_LONG_EDGE_KEY))
					return true;
			}
		}

		return longEdges;
	}

	protected AggregateItem getAggregate(int layer) {
		while (aggregateTable.getRowCount() <= layer) {
			aggregateTable.addItem();
		}

		return (AggregateItem) aggregateTable.getTuple(layer);
	}

	@SuppressWarnings("unchecked")
	protected int getMaxLayer(TupleSet ts) {
		Iterator<Tuple> iter = ts.tuples(null,
				new Sort(new String[] { PNConstants.NODE_LAYER_KEY }, new boolean[] { false }));
		if (iter.hasNext()) {
			return iter.next().getInt(PNConstants.NODE_LAYER_KEY);
		}

		return 0;
	}

	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		if (!enabled)
			aggregateTable.clear();
	}
}
