package org.processmining.humbertogarciacaballero.models.gui.components;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.util.Map;
import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.table.DefaultTableModel;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.prabhakardixit.processanalytics.ProcessAnalyticsViz;

import com.google.common.collect.Maps;

import prefuse.visual.NodeItem;

public class PlaceConformanceInformationPopUp extends JPanel {

	private NodeItem place;
	private JPanel panel;
	private JPanel panel1;
	private JTable table;
	private JTable table_1;
	private DefaultTableModel singleModel;
	private DefaultTableModel severalModel;

	/**
	 * Create the panel.
	 */
	public PlaceConformanceInformationPopUp() {
		setSize(new Dimension(400, 300));
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		panel = new JPanel();
		panel.setBorder(null);
		add(panel);
		panel.setLayout(new BorderLayout(0, 0));

		table = new JTable();
		singleModel = new DefaultTableModel(new Object[][] {}, new String[] { "Activity", "Frequency" }) {
			Class[] columnTypes = new Class[] { String.class, Float.class };

			public Class getColumnClass(int columnIndex) {
				return columnTypes[columnIndex];
			}
		};
		table.setModel(singleModel);
		table.getColumnModel().getColumn(0).setPreferredWidth(250);
		table.getColumnModel().getColumn(0).setMinWidth(250);
		panel.add(new JScrollPane(table), BorderLayout.CENTER);

		panel1 = new JPanel();
		panel1.setBorder(null);
		add(panel1);
		panel1.setLayout(new BorderLayout(0, 0));

		table_1 = new JTable();
		severalModel = new DefaultTableModel(new Object[][] {}, new String[] { "Activity", "Frequency" }) {
			Class[] columnTypes = new Class[] { String.class, Integer.class };

			public Class getColumnClass(int columnIndex) {
				return columnTypes[columnIndex];
			}
		};
		table_1.setModel(severalModel);
		table_1.getColumnModel().getColumn(0).setPreferredWidth(250);
		table_1.getColumnModel().getColumn(0).setMinWidth(250);
		panel1.add(new JScrollPane(table_1), BorderLayout.CENTER);
	}

	public void populatePopUp(NodeItem place) {
		clear();
		PNetVisualizerController controller = PNetVisualizerController.getInstance();

		// Look for the place object
		Place placeObject = null;
		for (PetrinetNode node : controller.getProcessAnalytics().getPetrinet().getNodes()) {
			if (node.getLocalID().localNodeID().toString().equals(place.getString(PNConstants.NODE_ID_KEY))) {
				placeObject = (Place) node;
			}
		}

		// Populate panel
		Map<String, Float> single = calculateSingleFrequency(placeObject);
		Map<String, Float> several = calculateSeveralFrequency(placeObject);

		for (String activity : single.keySet()) {
			singleModel.addRow(new Object[] { activity, single.get(activity) });
		}

		for (String activity : several.keySet()) {
			severalModel.addRow(new Object[] { activity, several.get(activity) });
		}

		// Store temporally
		this.place = place;

		invalidate();
		repaint();
	}

	private Map<String, Float> calculateSingleFrequency(Place place) {
		Map<String, Float> map = Maps.newHashMap();

		for (Set<Place> set : ProcessAnalyticsViz.logMovesFrequencyCurrent.keySet()) {
			if (set.contains(place) && set.size() == 1) {
				Map<String, Float> aux = ProcessAnalyticsViz.logMovesFrequencyCurrent.get(set);
				for (String activity : aux.keySet()) {
					if (map.containsKey(activity)) {
						map.put(activity, map.get(activity) + aux.get(activity));
					} else {
						map.put(activity, aux.get(activity));
					}
				}
			}
		}

		return map;
	}

	private Map<String, Float> calculateSeveralFrequency(Place place) {
		Map<String, Float> map = Maps.newHashMap();

		for (Set<Place> set : ProcessAnalyticsViz.logMovesFrequencyCurrent.keySet()) {
			if (set.contains(place) && set.size() > 1) {
				Map<String, Float> aux = ProcessAnalyticsViz.logMovesFrequencyCurrent.get(set);
				for (String activity : aux.keySet()) {
					if (map.containsKey(activity)) {
						map.put(activity, map.get(activity) + aux.get(activity));
					} else {
						map.put(activity, aux.get(activity));
					}
				}
			}
		}

		return map;
	}

	public NodeItem getPlace() {
		return place;
	}

	public void clear() {
		singleModel.setRowCount(0);
		severalModel.setRowCount(0);
		place = null;
	}
}
