package org.processmining.humbertogarciacaballero.models.gui.components;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Iterator;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JPanel;

import org.processmining.humbertogarciacaballero.models.gui.GraphVisualization;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import com.google.common.collect.Sets;
import com.google.common.collect.Sets.SetView;

import prefuse.Visualization;
import prefuse.data.tuple.TupleSet;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class TransitionHotBoxComponent extends NodeHotboxComponent {
	private int forwardDepth = 0;
	
	/**
	 * Create the frame.
	 */
	public TransitionHotBoxComponent(Visualization vis) {
		super(vis);
		initGui();
	}

	private void initGui() {
		setBackground(new Color(0, 0, 0, 0));
		setSize(new Dimension(309, 132));

		setLayout(new BorderLayout());

		JPanel panel = new JPanel();
		panel.setBackground(new Color(0, 0, 0, 0));
		add(panel, BorderLayout.SOUTH);
		JButton button = new JButton(new CustomCloseAction());
		panel.add(button);

		JPanel panel_1 = new JPanel();
		panel_1.setBackground(new Color(0, 0, 0, 0));
		add(panel_1, BorderLayout.NORTH);
		//        JButton button_1 = new JButton("OPTION");
		//        panel_1.add(button_1);

		JPanel panel_2 = new JPanel();
		panel_2.setBackground(new Color(0, 0, 0, 0));
		add(panel_2, BorderLayout.EAST);
		final JButton button_2 = new JButton(new ForwardAction("-1", false));
		button_2.addMouseListener(new MouseListener() {
			private MouseMotionListener l;

			@Override
			public void mouseReleased(MouseEvent e) {
				System.out.println("Mouse released");
				button_2.removeMouseMotionListener(l);
			}

			@Override
			public void mousePressed(MouseEvent e) {
				System.out.println("Mouse pressed");
				button_2.addMouseMotionListener(l = new MouseMotionListener() {

					@Override
					public void mouseMoved(MouseEvent e) {
					}

					@Override
					public void mouseDragged(MouseEvent e) {
						System.out.println("Mouse dragged");

					}
				});
			}

			@Override
			public void mouseExited(MouseEvent e) {
			}

			@Override
			public void mouseEntered(MouseEvent e) {
			}

			@Override
			public void mouseClicked(MouseEvent e) {
			}
		});
		button_2.addMouseMotionListener(new MouseMotionListener() {

			@Override
			public void mouseMoved(MouseEvent e) {
				// TODO Auto-generated method stub

			}

			@Override
			public void mouseDragged(MouseEvent e) {
				// TODO Auto-generated method stub

			}
		});
		panel_2.add(button_2);
		
		JButton btnNewButton = new JButton(new ForwardAction("0", null));
		panel_2.add(btnNewButton);
		
		JButton button_1 = new JButton(new ForwardAction("+1", true));
		panel_2.add(button_1);

		JPanel panel_3 = new JPanel();
		panel_3.setBackground(new Color(0, 0, 0, 0));
		add(panel_3, BorderLayout.WEST);
		JButton button_3 = new JButton(new BackwardAction());
		panel_3.add(button_3);
	}

	private class ForwardAction extends AbstractAction {
		private Boolean increment;
		private TupleSet greenGroup; 
		private TupleSet redGroup;

		public ForwardAction(String text, Boolean increment) {
			super(text);
			greenGroup = vis.getFocusGroup(GraphVisualization.GRAPH_NODES_FORWARD_INTERACTION_FOCUS_GROUP);
			redGroup = vis.getFocusGroup(GraphVisualization.GRAPH_NODES_MISSING_PLACE_FOCUS_GROUP);
			this.increment = increment;
		}

		public void actionPerformed(ActionEvent e) {
			if (increment == null) forwardDepth = 0;
			else if (increment) forwardDepth++;
			else forwardDepth = Math.max(0, forwardDepth - 1);

        	vis.setValue(GraphVisualization.GRAPH_NODES_GROUP, null, VisualItem.HIGHLIGHT, false);
        	greenGroup.clear();
        	redGroup.clear();
			
			Set<NodeItem> markNodes = Sets.newHashSet(item.inNeighbors());
			markTransitions(markNodes, item, forwardDepth);
			greenGroup.removeTuple(item);
			runAnimations();
		}

		private void markTransitions(Set<NodeItem> markNodes, NodeItem item, int depth) {
			if (depth >= 0) {
				// Highlight node
				item.setHighlighted(true);
				greenGroup.addTuple(item);

				// Mark missing places
				SetView<NodeItem> diff = Sets.difference(Sets.newHashSet(item.inNeighbors()), markNodes);
				for (NodeItem place : diff) {
					place.setHighlighted(true);
					redGroup.addTuple(place);
				}

				// Mark out neighbours
				markNodes.addAll(Sets.newHashSet(item.outNeighbors()));

				// Retrieve next transitions
				Iterator<NodeItem> iter = item.outNeighbors();
				while (iter.hasNext()) {
					NodeItem place = iter.next();
					Iterator<NodeItem> iter2 = place.outNeighbors();
					while (iter2.hasNext()) {
						markTransitions(markNodes, iter2.next(), depth - 1);
					}
				}
			}
		}
	}

	private class BackwardAction extends AbstractAction {
		public BackwardAction() {
			super("Backward");
		}

		public void actionPerformed(ActionEvent e) {
			System.out.println("println");
		}
	}
	
    class CustomCloseAction extends NodeHotboxComponent.CloseAction {
		private TupleSet greenGroup; 
		private TupleSet redGroup;
		
        public CustomCloseAction() {
            super();
			greenGroup = vis.getFocusGroup(GraphVisualization.GRAPH_NODES_FORWARD_INTERACTION_FOCUS_GROUP);
			redGroup = vis.getFocusGroup(GraphVisualization.GRAPH_NODES_MISSING_PLACE_FOCUS_GROUP);
        }

        public void actionPerformed(ActionEvent e) {
        	// Clean states
        	forwardDepth = 0;
        	vis.setValue(GraphVisualization.GRAPH_NODES_GROUP, null, VisualItem.HIGHLIGHT, false);
        	greenGroup.clear();
        	redGroup.clear();
        	super.actionPerformed(e);
        	runAnimations();
        }        
    }
    
    private void runAnimations() {
    	vis.cancel(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
		vis.run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
		vis.run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
    }
}
