package org.processmining.humbertogarciacaballero.models.gui.controllers;

import java.awt.event.MouseEvent;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Observable;
import java.util.Observer;
import java.util.Queue;
import java.util.Set;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.gui.GraphVisualization;
import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import com.google.common.collect.Sets;

import prefuse.Visualization;
import prefuse.controls.ControlAdapter;
import prefuse.data.Node;
import prefuse.data.tuple.TupleSet;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class PathControl extends ControlAdapter implements Observer {

    // Tables of elements used to
    private TupleSet incomingHighlight;
    private TupleSet outgoingHighlight;

    // Controller
    private PNetVisualizerController controller;

    // Current higlighted node
    private Set<NodeItem> highlightedNodes;
    private NodeType currentNodeType;
    private Visualization vis;

    // Path mode (incoming, outgoing, both)
    private PathMode pathMode;

    public static enum PathMode {
        INCOMING, OUTGOING, BOTH;
    }

    public PathControl(TupleSet incomingHighlihgt, TupleSet outgoingHighlight) {
        this.incomingHighlight = incomingHighlihgt;
        this.outgoingHighlight = outgoingHighlight;
        controller = PNetVisualizerController.getInstance();
        controller.addObserver(this);
        pathMode = controller.getPathMode();
        highlightedNodes = Sets.newHashSet();
    }

    /**
     * @see prefuse.controls.Control#itemClicked(prefuse.visual.VisualItem,
     *      java.awt.event.MouseEvent)
     */
    public void itemClicked(VisualItem item, MouseEvent e) {
        if (item instanceof NodeItem && e.getClickCount() <= 2) {
            NodeItem node = (NodeItem) item;
            handleClick(node, e.getClickCount());
        }
    }

    private void handleClick(NodeItem node, int clicks) {
        if (highlightedNodes.isEmpty()) {
            if (clicks == 2) {
                addNodeToCurrentState(node);
            }
        } else {
            if (highlightedNodes.contains(node)) {
                if (clicks == 1) {
                    removeNodeFromCurrentState(node);
                } else if (clicks == 2) {
                    resetState(node.getVisualization());
                    addNodeToCurrentState(node);
                }
            } else {
                if (clicks == 1) {
                    addNodeToCurrentState(node);
                } else {
                    resetState(node.getVisualization());
                    addNodeToCurrentState(node);
                }
            }
        }

        runAnimations();
    }

    private void removeNodeFromCurrentState(NodeItem node) {

    }

    private void resetState(Visualization vis) {
        vis.setValue(GraphVisualization.GRAPH_EDGES_GROUP, null, PNConstants.EDGE_SIMPLIFICABLE, true);
        vis.setValue(GraphVisualization.GRAPH_GROUP, null, VisualItem.HIGHLIGHT, false);
        incomingHighlight.clear();
        outgoingHighlight.clear();
        highlightedNodes.clear();
        currentNodeType = null;
        controller.setSliderEnabled(false);
    }

    private void addNodeToCurrentState(NodeItem node) {
        // Checks selection node type
        NodeType nodeType = NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY));
        if (highlightedNodes.isEmpty()) {
            currentNodeType = nodeType;
            vis = node.getVisualization();
        } else {
            if (!currentNodeType.equals(nodeType)) {
                JOptionPane.showMessageDialog(null, "The selected nodes are not of the same type.");
                return;
            }
        }

        // Adds node to the current set
        highlightedNodes.add(node);

        // Highlight node
        node.setHighlighted(true);

        // Shows path involving node
        showPathInvolvingNode(controller.getDepth());

        controller.setSliderEnabled(true);
        controller.setMaxDepth(50);
    }

    // private void handleDoubleClick(NodeItem node, boolean controlPressed) {
    // if (highlightedNodes.contains(node)) {
    // if (controlPressed) {
    // node.setHighlighted(false);
    // clearInformationPanel();
    // hideAllPaths(node);
    // highlightedNodes.remove(node);
    // if (highlightedNodes.isEmpty()) {
    // controller.resetDepth();
    // controller.setSliderEnabled(false);
    // }
    // } else {
    // for (NodeItem highlightedNode : highlightedNodes) {
    // highlightedNode.setHighlighted(false);
    // hideAllPaths(highlightedNode);
    // }
    // if (highlightedNodes.size() == 1) {
    // highlightedNodes.clear();
    // } else {
    // highlightNode(node);
    // }
    // }
    // } else {
    // // Highlighting
    // if (!controlPressed) {
    // for (NodeItem highlightedNode : highlightedNodes) {
    // highlightedNode.setHighlighted(false);
    // hideAllPaths(highlightedNode);
    // }
    // highlightedNodes.clear();
    // currentNodeType =
    // NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY));
    // } else if (highlightedNodes.isEmpty()) {
    // currentNodeType =
    // NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY));
    // } else if
    // (!NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)).equals(currentNodeType))
    // {
    // JOptionPane.showMessageDialog(null, "The selected nodes are not of the
    // same type.");
    // return;
    // }
    //
    // highlightNode(node);
    // }
    //
    // runAnimations(node.getVisualization());
    // }

    // private void highlightNode(NodeItem node) {
    // node.setHighlighted(true);
    // highlightedNodes.add(node);
    //
    // // Paths
    // showPathInvolvingNode(node, controller.getDepth());
    //
    // int maxDepth = calculateMaxDepth(node, pathMode);
    //
    // controller.setMaxDepth(maxDepth);
    // controller.setSliderEnabled(true);
    // }

    private void runAnimations() {
        vis.cancel(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
        // vis.cancel(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
        vis.run(PNConstants.VIS_FILL_NODES_ACTIONS_NAME);
        vis.run(PNConstants.VIS_FILL_NODES_DEPTH_ACTIONS_NAME);
        vis.run(PNConstants.VIS_TEXT_COLOR_ACTIONS_NAME);
        vis.run(PNConstants.VIS_FILL_EDGES_ACTIONS_NAME);
        vis.run(PNConstants.VIS_STROKE_EDGES_ACTIONS_NAME);
        // vis.run(PNConstants.VIS_LAYOUT_NO_PAINT_ACTIONS_NAME);
        // vis.run(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
        vis.run(PNConstants.VIS_HIGHLIGHT_ACTIONS_NAME);
    }

    // private void hideAllPaths(NodeItem item) {
    // item.getVisualization().setValue(GraphVisualization.GRAPH_EDGES_GROUP,
    // null, PNConstants.EDGE_SIMPLIFICABLE,
    // true);
    // item.getVisualization().setValue(GraphVisualization.GRAPH_GROUP, null,
    // VisualItem.HIGHLIGHT, false);
    // incomingHighlight.clear();
    // outgoingHighlight.clear();
    // }

    @SuppressWarnings("unchecked")
    private void showPathInvolvingNode(int depth) {
        incomingHighlight.clear();
        outgoingHighlight.clear();

        if (depth > 0) {
            if (NodeType.TRANSITION.equals(currentNodeType)) {
                // Set<NodeItem> preconditions, postconditions;
                // preconditions = Sets.newHashSet(node.inNeighbors());
                // findConditionsTransition(node, preconditions, depth);
                // postconditions = Sets.newHashSet(node.outNeighbors());
                // fireTransition(node, postconditions, depth);
            } else {
                // findTriggersPlace(node, depth);
                consumeTokenPlace(Sets.newHashSet(highlightedNodes), vis, depth);
            }
        }
    }

    // @SuppressWarnings("unchecked")
    // private void showPathInvolvingNode(NodeItem node, int depth) {
    // if (depth > 0) {
    // if
    // (NodeType.TRANSITION.equals(NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY))))
    // {
    // Set<NodeItem> preconditions, postconditions;
    // switch (pathMode) {
    // case INCOMING:
    // preconditions = Sets.newHashSet(node.inNeighbors());
    // findConditionsTransition(node, preconditions, controller.getDepth());
    // break;
    // case OUTGOING:
    // postconditions = Sets.newHashSet(node.outNeighbors());
    // fireTransition(node, postconditions, controller.getDepth());
    // break;
    // case BOTH:
    // preconditions = Sets.newHashSet(node.inNeighbors());
    // findConditionsTransition(node, preconditions, controller.getDepth());
    // postconditions = Sets.newHashSet(node.outNeighbors());
    // fireTransition(node, postconditions, controller.getDepth());
    // break;
    // }
    // } else {
    // switch (pathMode) {
    // case INCOMING:
    // findTriggersPlace(node, controller.getDepth());
    // ;
    // break;
    // case OUTGOING:
    // consumeTokenPlace(node, controller.getDepth());
    // break;
    // case BOTH:
    // findTriggersPlace(node, controller.getDepth());
    // consumeTokenPlace(node, controller.getDepth());
    // break;
    // }
    // }
    // }
    // }

    @SuppressWarnings("unchecked")
    private void fireTransition(NodeItem node, Set<NodeItem> marking, int depth) {
        if (depth > 0) {
            Visualization vis = node.getVisualization();

            for (NodeItem adjacent : marking) {
                // EdgeItem edge = iter.next();
                // edge.setBoolean(PNConstants.EDGE_SIMPLIFICABLE, false);
                // NodeItem adjacent = (NodeItem) edge.getAdjacentNode(node);
                EdgeItem e = (EdgeItem) node.getGraph().getEdge(node, adjacent);
                e.setHighlighted(true);
                outgoingHighlight.addTuple(e);
                adjacent.setHighlighted(true);
                outgoingHighlight.addTuple(adjacent);
            }

            Set<NodeItem> enabledTransitions = Sets.newHashSet();
            Iterator<NodeItem> nodes = vis.items(GraphVisualization.GRAPH_NODES_GROUP,
                    PNConstants.TRANSITION_NODE_PREDICATE);
            while (nodes.hasNext()) {
                NodeItem transition = nodes.next();
                if (marking.containsAll(Sets.newHashSet(transition.inNeighbors()))) {
                    enabledTransitions.add(transition);
                }
            }

            for (NodeItem transition : enabledTransitions) {
                transition.setHighlighted(true);
                outgoingHighlight.addTuple(transition);
                if (!incomingHighlight.containsTuple(transition)) {
                    Iterator<EdgeItem> iter = transition.inEdges();
                    while (iter.hasNext()) {
                        EdgeItem e = iter.next();
                        e.setHighlighted(true);
                        outgoingHighlight.addTuple(e);
                    }
                    Set<NodeItem> conditions = Sets.newHashSet(transition.inNeighbors());
                    Set<NodeItem> postconditions = Sets.newHashSet(transition.outNeighbors());
                    marking.removeAll(conditions);
                    marking.addAll(postconditions);
                    fireTransition(transition, marking, depth - 1);
                    marking.addAll(conditions);
                    marking.removeAll(postconditions);
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void findConditionsTransition(NodeItem node, Set<NodeItem> marking, int depth) {
        if (depth > 0) {
            Iterator<EdgeItem> iter = node.inEdges();
            while (iter.hasNext()) {
                EdgeItem e = iter.next();
                e.setHighlighted(true);
                incomingHighlight.addTuple(e);
                NodeItem neighbour = e.getAdjacentItem(node);
                neighbour.setHighlighted(true);
                incomingHighlight.addTuple(neighbour);
                Iterator<NodeItem> iter2 = neighbour.inNeighbors();
                while (iter2.hasNext()) {
                    NodeItem transition = iter2.next();
                    e = (EdgeItem) neighbour.getGraph().getEdge(transition, neighbour);
                    e.setHighlighted(true);
                    incomingHighlight.addTuple(e);
                    transition.setHighlighted(true);
                    incomingHighlight.addTuple(transition);
                    if (!outgoingHighlight.containsTuple(transition)) {
                        findConditionsTransition(transition, marking, depth - 1);
                    }
                }
            }
        }
    }

    private void findTriggersPlace(NodeItem place, int depth) {
        if (depth > 0) {
            Iterator<NodeItem> iter = place.inNeighbors();
            while (iter.hasNext()) {
                NodeItem transition = iter.next();
                EdgeItem e = (EdgeItem) place.getGraph().getEdge(transition, place);
                e.setHighlighted(true);
                incomingHighlight.addTuple(e);
                Set<NodeItem> marking = Sets.newHashSet(transition.inNeighbors());
                incomingHighlight.addTuple(transition);
                transition.setHighlighted(true);
                findConditionsTransition(transition, marking, depth - 1);
            }
        }
    }

    private void consumeTokenPlace(Set<NodeItem> marking, Visualization vis, int depth) {
        if (depth > 0) {
            // Finds enabled transitions
            Set<NodeItem> enabledTransitions = Sets.newHashSet();
            Iterator<NodeItem> nodes = vis.items(GraphVisualization.GRAPH_NODES_GROUP,
                    PNConstants.TRANSITION_NODE_PREDICATE);
            while (nodes.hasNext()) {
                NodeItem transition = nodes.next();
                if (marking.containsAll(Sets.newHashSet(transition.inNeighbors()))) {
                    enabledTransitions.add(transition);
                }
            }

            for (NodeItem transition : enabledTransitions) {
                transition.setHighlighted(true);
                outgoingHighlight.addTuple(transition);
                if (!incomingHighlight.containsTuple(transition)) {
                    Iterator<EdgeItem> iter = transition.inEdges();
                    while (iter.hasNext()) {
                        EdgeItem e = iter.next();
                        e.setHighlighted(true);
                        outgoingHighlight.addTuple(e);
                    }
                    iter = transition.outEdges();
                    while (iter.hasNext()) {
                        EdgeItem e = iter.next();
                        e.setHighlighted(true);
                        outgoingHighlight.addTuple(e);
                    }
                    Set<NodeItem> conditions = Sets.newHashSet(transition.inNeighbors());
                    Set<NodeItem> postconditions = Sets.newHashSet(transition.outNeighbors());
                    for (NodeItem postcondition : postconditions) {
                        postcondition.setHighlighted(true);
                        outgoingHighlight.addTuple(postcondition);
                    }
                    marking.removeAll(conditions);
                    marking.addAll(postconditions);
                    consumeTokenPlace(marking, vis, depth - 1);
                    marking.addAll(conditions);
                    marking.removeAll(postconditions);
                }
            }
        }
    }

    // private void consumeTokenPlace(Node place, int depth) {
    // if (depth > 0) {
    // Iterator<NodeItem> iter = place.outNeighbors();
    // while (iter.hasNext()) {
    // NodeItem transition = iter.next();
    // Set<NodeItem> preconditions = Sets.newHashSet(transition.inNeighbors());
    // if (preconditions.size() == 1 && preconditions.contains(place)) {
    // EdgeItem e = (EdgeItem) place.getGraph().getEdge(place, transition);
    // e.setHighlighted(true);
    // outgoingHighlight.addTuple(e);
    // Set<NodeItem> marking = Sets.newHashSet(transition.outNeighbors());
    // outgoingHighlight.addTuple(transition);
    // transition.setHighlighted(true);
    // fireTransition(transition, marking, depth - 1);
    // }
    // }
    // }
    // }

    @SuppressWarnings("unchecked")
    private int calculateMaxDepth(NodeItem node, PathMode pathMode) {
        int maxDepth = 0;
        node.getVisualization().setValue(GraphVisualization.GRAPH_NODES_GROUP, null, PNConstants.NODE_DEPTH_KEY, -1);
        Queue<Node> queue = new LinkedList<Node>();
        queue.add(node);
        node.setInt(PNConstants.NODE_DEPTH_KEY, 0);

        while (!queue.isEmpty()) {
            Node current = queue.poll();
            Iterator<Node> iter;

            switch (pathMode) {
            case INCOMING:
                iter = current.inNeighbors();
                break;
            case OUTGOING:
                iter = current.outNeighbors();
                break;
            default:
                iter = current.neighbors();
                break;
            }

            while (iter.hasNext()) {
                Node neighbor = iter.next();
                if (neighbor.getInt(PNConstants.NODE_DEPTH_KEY) == -1) {
                    int depth = current.getInt(PNConstants.NODE_DEPTH_KEY) + 1;
                    neighbor.setInt(PNConstants.NODE_DEPTH_KEY, depth);
                    maxDepth = Math.max(maxDepth, depth);
                    queue.add(neighbor);
                }
            }
        }

        return maxDepth;
    }

    public void clearInformationPanel() {
        JPanel panel = controller.getInfoPanel();
        panel.removeAll();
        panel.revalidate();
        panel.repaint();
    }

    @Override
    public void update(Observable observable, Object value) {
        if (observable instanceof PNetVisualizerController && value instanceof Integer) {
            showPathInvolvingNode(controller.getDepth());
            runAnimations();
        }
        // } else if (observable instanceof PNetVisualizerController && value
        // instanceof PathMode) {
        // if (pathMode != controller.getPathMode()) {
        // pathMode = controller.getPathMode();
        // controller.setMaxDepth(calculateMaxDepth(highlightedNodes,
        // pathMode));
        // }
        //
        // changePathAndHighlights();
        // }
    }

    // private void changePathAndHighlights() {
    // if (highlightedNodes != null) {
    // // Paths
    // hideAllPaths(highlightedNodes);
    // highlightedNodes.setHighlighted(true);
    // showPathInvolvingNode(highlightedNodes, controller.getDepth());
    // runAnimations(highlightedNodes.getVisualization());
    // }
    // }
}
