package org.processmining.humbertogarciacaballero.models.model.commands;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.gui.GraphVisualization;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayoutOrientation;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LineType;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.emplacer.EmplacingStrategy;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.layerer.LayeringStrategy;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.sequencer.SequencingStrategy;

import prefuse.action.Action;
import prefuse.activity.Activity;
import prefuse.activity.ActivityAdapter;

public class ChangeParametersCommand extends AbstractCommand {

    private GraphVisualization vis;
    private PNetVisualizerController controller;

    // Layout parameters
    private boolean firstStagesEnabled;
    private LayeringStrategy layeringStrategy;
    private LayoutOrientation layoutOrientation;
    private boolean enabledSimplifier;
    private boolean enabledSimplifyAll;
    private boolean enabledSimplifyLongEdges;
    private boolean enabledEdgeDirection;
    private boolean enabledBarDrawing;
    private boolean enabledSimplifySameLayer;
    private boolean enabledReorderNodes;
    private int crossingThreshold;
    private int totalCrossingThreshold;
    private int maxLoops;
    private int intracellSpacing;
    private int layerSpace;
    private float edgeLandingOffset;
    private SequencingStrategy sequencingStrategy;
    private EmplacingStrategy emplacingStrategy;
    private LineType lineType;

    // Graph visualization parameters
    private int transitionsWidth;
    private int transitionsHeight;
    private boolean transitionsInOutDegree;
    private int placesSize;
    private int fontSize;
    private int degreeFontSize;
    private long animationsLength;
    
    private int edgeColor;
    private int strokeColor;

    public ChangeParametersCommand(GraphVisualization vis, boolean firstStagesEnabled,
            LayeringStrategy layeringStrategy, LayoutOrientation layoutOrientation, boolean enabledSimplifier,
            boolean enabledSimplifyAll, boolean enabledSimplifyLongEdges, boolean enabledEdgeDirection,
            boolean enabledBarDrawing, boolean enabledSimplifySameLayer, boolean enabledReorderNodes, int crossingThreshold,
            int totalCrossingThreshold, int maxLoops, int intracellSpacing, int layerSpace, float edgeLandingOffset,
            SequencingStrategy sequencingStrategy, EmplacingStrategy emplacingStrategy, LineType lineType,
            int transitionsWidth, int transitionsHeight, boolean transitionsInOutDegree, int placesSize, int placesHeight, int fontSize,
            int degreeFontSize, long animationLength, int edgeColor, int strokeColor) {
        this.vis = vis;
        this.controller = PNetVisualizerController.getInstance();

        // Layout parameters
        this.firstStagesEnabled = firstStagesEnabled;
        this.layeringStrategy = layeringStrategy;
        this.layoutOrientation = layoutOrientation;
        this.enabledSimplifier = enabledSimplifier;
        this.enabledSimplifyAll = enabledSimplifyAll;
        this.enabledSimplifyLongEdges = enabledSimplifyLongEdges;
        this.enabledEdgeDirection = enabledEdgeDirection;
        this.enabledBarDrawing = enabledBarDrawing;
        this.enabledSimplifySameLayer = enabledSimplifySameLayer;
        this.enabledReorderNodes = enabledReorderNodes;
        this.crossingThreshold = crossingThreshold;
        this.totalCrossingThreshold = totalCrossingThreshold;
        this.maxLoops = maxLoops;
        this.intracellSpacing = intracellSpacing;
        this.layerSpace = layerSpace;
        this.edgeLandingOffset = edgeLandingOffset;
        this.sequencingStrategy = sequencingStrategy;
        this.emplacingStrategy = emplacingStrategy;
        this.lineType = lineType;

        // Visualization parameters
        this.transitionsWidth = transitionsWidth;
        this.transitionsHeight = transitionsHeight;
        this.transitionsInOutDegree = transitionsInOutDegree;
        this.placesSize = placesSize;
        this.fontSize = fontSize;
        this.degreeFontSize = degreeFontSize;
        this.animationsLength = animationLength;
        
        this.edgeColor = edgeColor;
        this.strokeColor = strokeColor;
    }

    @Override
    public void doIt() {
        updateLayoutParameters();
        updateVisualizationParameters();
        executeLayout();
    }

    @Override
    public void undoIt() {
        updateLayoutParameters();
        updateVisualizationParameters();
        executeLayout();
    }

    private void updateLayoutParameters() {
        boolean booleanAux = controller.isFirstStagesEnabled();
        controller.setFirstStagesEnabled(firstStagesEnabled);
        firstStagesEnabled = booleanAux;

        LayeringStrategy oldLayeringStrategy = controller.getLayeringStrategy();
        controller.setLayeringStrategy(layeringStrategy);
        layeringStrategy = oldLayeringStrategy;

        LayoutOrientation oldLayoutOrientation = controller.getLayoutOrientation();
        controller.setLayoutOrientation(layoutOrientation);
        layoutOrientation = oldLayoutOrientation;

        booleanAux = controller.isSimplifierEnabled();
        controller.setSimplifierEnable(enabledSimplifier);
        enabledSimplifier = booleanAux;

        booleanAux = controller.isSimplifyShortEdgesEnabled();
        controller.setSimplifyShortEdgesEnabled(enabledSimplifyAll);
        enabledSimplifyAll = booleanAux;
        
        booleanAux = controller.isSimplifyLongEdgesEnabled();
        controller.setSimplifyLongEdgesEnabled(enabledSimplifyLongEdges);
        enabledSimplifyLongEdges = booleanAux;

        booleanAux = controller.isEdgeDirectionEnabled();
        controller.setEdgeDirectionEnabled(enabledEdgeDirection);
        enabledEdgeDirection = booleanAux;

        booleanAux = controller.isBarDrawingEnabled();
        controller.setBarDrawingEnabled(enabledBarDrawing);
        enabledBarDrawing = booleanAux;

        booleanAux = controller.isSimplifySameLayer();
        controller.setSimplifySameLayer(enabledSimplifySameLayer);
        enabledSimplifySameLayer = booleanAux;
        
        booleanAux = controller.isReorderNodesEnabled();
        controller.setReorderNodesEnabled(enabledReorderNodes);
        enabledReorderNodes = booleanAux;

        int intAux = controller.getSimplifierCrossingThreshold();
        controller.setSimplifierCrossingThreshold(crossingThreshold);
        crossingThreshold = intAux;

        intAux = controller.getSimplifierTotalCrossingThreshold();
        controller.setSimplifierTotalCrossingThreshold(totalCrossingThreshold);
        totalCrossingThreshold = intAux;

        intAux = controller.getSimplifierMaxLoops();
        controller.setSimplifierMaxLoops(maxLoops);
        maxLoops = intAux;

        intAux = controller.getIntracellSpacing();
        controller.setIntracellSpacing(intracellSpacing);
        intracellSpacing = intAux;

        intAux = controller.getLayerSpacing();
        controller.setLayerSpacing(layerSpace);
        layerSpace = intAux;

        float floatAux = controller.getEdgeLandingOffset();
        controller.setEdgeLandingOffset(edgeLandingOffset);
        edgeLandingOffset = floatAux;

        SequencingStrategy oldSequencingStrategy = controller.getSequencingStrategy();
        controller.setSequencingStrategy(sequencingStrategy);
        sequencingStrategy = oldSequencingStrategy;

        EmplacingStrategy oldEmplacingStrategy = controller.getEmplacingStrategy();
        controller.setEmplacingStrategy(emplacingStrategy);
        emplacingStrategy = oldEmplacingStrategy;

        LineType oldLineType = controller.getLineType();
        controller.setLineType(lineType);
        lineType = oldLineType;
    }

    private void updateVisualizationParameters() {
        int aux = vis.getTransitionsWidth();
        vis.setTransitionsWidth(transitionsWidth);
        transitionsWidth = aux;

        aux = vis.getTransitionsHeight();
        vis.setTransitionsHeight(transitionsHeight);
        transitionsHeight = aux;
        
        boolean booleanAux = vis.isTransitionInOutDegreeEnabled();
        vis.setTransitionInOutDegree(transitionsInOutDegree);
        transitionsInOutDegree = booleanAux;

        aux = vis.getPlacesSize();
        vis.setPlacesSize(placesSize);
        placesSize = aux;

        aux = vis.getFontSize();
        vis.setFontSize(fontSize);
        fontSize = aux;

        aux = vis.getDegreeFontSize();
        vis.setDegreeFontSize(degreeFontSize);
        degreeFontSize = aux;

        long animation = vis.getAnimationLength();
        vis.setAnimationLength(animationsLength);
        animationsLength = animation;
        
        aux = vis.getEdgeDefaultColor();
        vis.setEdgeDefaultColor(edgeColor);
        edgeColor = aux;
        
        aux = vis.getShapeBorderColor();
        vis.setShapeBorderColor(strokeColor);
        strokeColor = aux;
    }

    private void executeLayout() {
        final Action a = vis.getVisualization().getAction(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
        a.addActivityListener(new ActivityAdapter() {
            public void activityCancelled(Activity activity) {
                vis.getVisualization().getAction(PNConstants.VIS_LAYOUT_ACTIONS_NAME).run(1.0);
                finish();
                a.removeActivityListener(this);
            }

            public void activityFinished(Activity activity) {
                finish();
                a.removeActivityListener(this);
            }
        });

        vis.getVisualization().cancel(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
        vis.getVisualization().run(PNConstants.VIS_LAYOUT_NO_PAINT_ACTIONS_NAME);
        vis.getVisualization().run(PNConstants.VIS_ANIMATE_LAYOUT_ACTIONS_NAME);
    }
}
