package org.processmining.humbertogarciacaballero.models.model.layout.stage.layerer;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;

import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import prefuse.data.Graph;
import prefuse.data.Node;

public abstract class BFSLayerer extends Layerer {

	private boolean undirected;
	
	public BFSLayerer(HierarchicalLayoutContext context, boolean undirected) {
		super(context);
		this.undirected = undirected;
	}
	
	@Override
	public void calculateLayers(Graph graph) {
		Iterator<Node> rootNodes = getRootNodes(graph);
		
		if (rootNodes.hasNext()) {
			Queue<Node> queue = new LinkedList<>();
			
			while (rootNodes.hasNext()) {
				Node node = rootNodes.next();
				setLayer(node, 0);
				queue.add(node);
			}
			
			while (!queue.isEmpty()) {
				Node current = queue.poll();
				Iterator<Node> neighbors = getSuccessors(current, undirected);
				
				while (neighbors.hasNext()) {
					Node target = neighbors.next();
					if (getLayer(target) == -1) {
						setLayer(target, getLayer(current) + 1);
						queue.add(target);
					} else {
						createDummies(graph, current, target);	
					}
				}
			}
		}
	}

	public boolean isUndirected() {
		return undirected;
	}
	
	public void setUndirected(boolean undirected) {
		this.undirected = undirected;
	}
}
