package org.processmining.humbertogarciacaballero.models.model.layout.stage.layerer;

import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.EdgeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.NodeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.LayoutStage;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public abstract class Layerer extends LayoutStage {

    public Layerer(HierarchicalLayoutContext context) {
        super(context);
    }

    @Override
    public int run(Graph graph) {
        reset();
        calculateLayers(graph);

        return 0;
    }

    public abstract void calculateLayers(Graph graph);

    protected void setLayer(Node node, int layer) {
        getContext().getLayerContainer().addNode(node, layer);
    }

    protected void moveNode(Node n, int newLayer) {
        getContext().getLayerContainer().moveNode(n, newLayer);
    }

    public int getLayer(Node node) {
        return getContext().getLayerContainer().getLayer(node);
    }

    protected void reset() {
        getContext().getLayerContainer().reset();
    }

    protected List<Node> getNodes(int layer) {
        return getContext().getLayerContainer().getNodes(layer);
    }

    public void createDummies(Graph graph, Node current, Node target) {
        createDummies(graph, current, target, graph.getEdge(current, target));
        createDummies(graph, target, current, graph.getEdge(target, current));
    }

    private void createDummies(Graph graph, Node source, Node target, Edge e) {
        int difference;
        if ((difference = Math.abs(getLayer(source) - getLayer(target))) > 1 && e != null) {
            EdgeParams edgeParams = (EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS);
            Node nodeAux = source;
            for (int i = 1; i < difference; i++) {
                Node dummyNode = createDummyNode(graph, source, target);

                Edge dummyEdge = graph.addEdge(nodeAux, dummyNode);
                dummyEdge.set(HierarchicalLayouter.EDGE_PARAMS, edgeParams.clone());

                if (getLayer(nodeAux) > getLayer(target)) {
                    setLayer(dummyNode, getLayer(nodeAux) - 1);
                } else {
                    setLayer(dummyNode, getLayer(nodeAux) + 1);
                }
                nodeAux = dummyNode;
            }
            Edge dummyEdge = graph.addEdge(nodeAux, target);
            dummyEdge.set(HierarchicalLayouter.EDGE_PARAMS, edgeParams.clone());
            graph.removeEdge(e);
        }
    }

    private Node createDummyNode(Graph graph, Node source, Node target) {
        Node n = graph.addNode();
        NodeParams params = (NodeParams) n.get(HierarchicalLayouter.NODE_PARAMS);
        NodeParams sourceParams = (NodeParams) source.get(HierarchicalLayouter.NODE_PARAMS);

        params.dummy = true;
        params.width = 1;
        params.height = 1;
        sourceParams.addDummy(target, n);

        return n;
    }
}
