package org.processmining.humbertogarciacaballero.models.model.layout.stage.lines;

import java.awt.geom.Point2D;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.EdgeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayoutOrientation;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import prefuse.data.Edge;
import prefuse.data.Node;

public class PolylineBendCalculator extends BendCalculator {

    private int dummySize;

    public PolylineBendCalculator(HierarchicalLayoutContext context) {
        super(context);
    }

    @Override
    public void createBends(Edge e, int layerDifference, List<Node> dummies) {
        LayoutOrientation orientation = getContext().getLayoutOrientation();
        setDummySize(orientation);
        EdgeParams params = ((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS));

        List<Point2D> bends = params.bends;

        if (dummies != null) {
            for (Node dummy : dummies) {
                bends.add(
                        calculateIntermediatePoint(getX(dummy), getY(dummy), orientation.opposite(), layerDifference));
                bends.add(new Point2D.Double(getX(dummy), getY(dummy)));
                bends.add(calculateIntermediatePoint(getX(dummy), getY(dummy), orientation, layerDifference));
            }
        }
    }

    private void setDummySize(LayoutOrientation orientation) {
        switch (orientation) {
        case TOP_TO_BOTTOM:
        case BOTTOM_TO_TOP:
            dummySize = Math.max(PNetVisualizerController.getInstance().getPlacesSize(), PNetVisualizerController.getInstance().getTransitionsHeight());
            break;
        default:
            dummySize = Math.max(PNetVisualizerController.getInstance().getPlacesSize(), PNetVisualizerController.getInstance().getTransitionsWidth());
            break;
        }
        dummySize /= 2;
    }

    protected Point2D calculateIntermediatePoint(double x, double y, LayoutOrientation orientation,
            int layerDifference) {
        float edgeLandingOffset = getContext().getEdgeLandingOffset();
        Point2D p;
        switch (layerDifference <= 0 ? orientation : orientation.opposite()) {
        case LEFT_TO_RIGHT:
            p = new Point2D.Double(x + dummySize + edgeLandingOffset, y);
            break;
        case RIGHT_TO_LEFT:
            p = new Point2D.Double(x - dummySize - edgeLandingOffset, y);
            break;
        case TOP_TO_BOTTOM:
            p = new Point2D.Double(x, y + dummySize + edgeLandingOffset);
            break;
        default:
            p = new Point2D.Double(x, y - dummySize - edgeLandingOffset);
            break;
        }

        return p;
    }
}
