package org.processmining.humbertogarciacaballero.plugins;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.humbertogarciacaballero.algorithms.ProcessMiningAlgorithms;
import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.SoundnessInformation;
import org.processmining.humbertogarciacaballero.models.util.Utils;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.InhibitorArc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.ResetArc;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Tuple;
import prefuse.data.expression.ColumnExpression;
import prefuse.data.expression.ComparisonPredicate;
import prefuse.data.expression.ObjectLiteral;

public class PluginUtils {

	@SuppressWarnings("unchecked")
	public static Graph netToGraph(PluginContext context, AcceptingPetriNet n, SoundnessInformation information) {
		Graph graph = netToGraph(context, n.getNet(), null, information);
		Iterator<Place> iter = n.getInitialMarking().iterator();
		while (iter.hasNext()) {
			Place p = iter.next();
			Iterator<Tuple> iter2 = graph.getNodes().tuples(new ComparisonPredicate(ComparisonPredicate.EQ,
					new ColumnExpression(PNConstants.NODE_ID_KEY), new ObjectLiteral(p.getLocalID().toString())));

			if (iter2.hasNext()) {
				iter2.next().set(PNConstants.NODE_STARTING_KEY, true);
			}
		}

		Iterator<Marking> markings = n.getFinalMarkings().iterator();
		while (markings.hasNext()) {
			iter = markings.next().iterator();
			while (iter.hasNext()) {
				Place p = iter.next();
				Iterator<Tuple> iter2 = graph.getNodes().tuples(new ComparisonPredicate(ComparisonPredicate.EQ,
						new ColumnExpression(PNConstants.NODE_ID_KEY), new ObjectLiteral(p.getLocalID().toString())));

				if (iter2.hasNext()) {
					iter2.next().set(PNConstants.NODE_ENDING_KEY, true);
				}
			}
		}
		return graph;
	}

	public static Graph netToGraph(PluginContext context, PetrinetGraph d, Marking m,
			SoundnessInformation information) {
		Graph graph = new Graph(true);
		Map<PetrinetNode, Integer> nodeMap = new HashMap<>();
		Utils.initSchema(graph);
		Map<String, String> labels = ProcessMiningAlgorithms.preprocessDiagnosisPetrinet(context, (Petrinet) d);
//		WoflanDiagnosis diagnosis = ProcessMiningAlgorithms.diagnosePetrinet(context, (Petrinet) d);
//		System.out.println(diagnosis.toHTMLString(true));

//		information.setOptionToComplete(diagnosis.isLive());
//		information.setProperCompletion(diagnosis.isBounded());
//		information.setNoDeadTasks(diagnosis.isNotDead());

		int V = d.getNodes().size();
		PetrinetNode[] nodes = d.getNodes().toArray(new PetrinetNode[V]);

		for (PetrinetNode node : nodes) {
			Node n = graph.addNode();
			nodeMap.put(node, n.getRow());
			if (node instanceof Place) {
				n.set(PNConstants.NODE_TYPE_KEY, NodeType.PLACE.toString());
				n.set(PNConstants.NODE_LABEL_KEY, labels.get(node.getId().toString()));

				// Diagnosis information
//				if (diagnosis.isBounded() == Boolean.FALSE) {
//					for (Place p : diagnosis.getUnboundedPlaces()) {
//						if (p.getLabel().equals(node.getLabel())) {
//							n.set(PNConstants.NODE_UNBOUNDED_PLACE_KEY, true);
//							information.addUnboundedPlace(node.getLocalID().localNodeID().toString());
//							break;
//						}
//					}
//				}
			} else {
				n.set(PNConstants.NODE_TYPE_KEY, NodeType.TRANSITION.toString());
				boolean tau;
				n.set(PNConstants.NODE_TAU_TRANSITION_KEY, (tau = ((Transition) node).isInvisible()));

				if (tau) {
					n.set(PNConstants.NODE_LABEL_KEY, "TAU");
				} else {
					n.set(PNConstants.NODE_LABEL_KEY, labels.get(node.getId().toString()));
				}

				// Diagnosis information
//				if (diagnosis.isLive() == Boolean.FALSE) {
//					for (Transition t : diagnosis.getNonLiveTransitions()) {
//						if (t.getLabel().equals(node.getLabel())) {
//							n.set(PNConstants.NODE_NON_LIVE_TRANSITION_KEY, true);
//							information.addNonLiveTransition(node.getLocalID().localNodeID().toString());
//							break;
//						}
//					}
//				}
//				if (diagnosis.isNotDead() == Boolean.FALSE) {
//					for (Transition t : diagnosis.getDeadTransitions()) {
//						if (t.getLabel().equals(node.getLabel())) {
//							n.set(PNConstants.NODE_DEAD_TRANSITION_KEY, true);
//							information.addDeadTransition(node.getLocalID().localNodeID().toString());
//							break;
//						}
//					}
//				}
			}
			n.setString(PNConstants.NODE_ID_KEY, node.getLocalID().localNodeID().toString());
		}

		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : d.getEdges()) {
			int source = nodeMap.get(edge.getSource());
			int target = nodeMap.get(edge.getTarget());
			Edge e = graph.getEdge(graph.addEdge(source, target));
			if (edge instanceof ResetArc) {
				e.setBoolean(PNConstants.EDGE_RESET_KEY, true);
			} else if (edge instanceof InhibitorArc) {
				e.setBoolean(PNConstants.EDGE_INHIBITOR_KEY, true);
			}
			if (d.getArc(edge.getTarget(), edge.getSource()) != null) {
				e.setBoolean(PNConstants.EDGE_BIDIRECTIONAL, true);
				if (graph.getEdge(target, source) != -1) {
					e.setBoolean(PNConstants.EDGE_BIDIRECTIONAL_DRAW, false);
				}
			}
		}

		// Sequences of diagnosis
//		if (diagnosis.isLive() == Boolean.FALSE) {
//			for (Pair<Marking, Transition> pair : diagnosis.getNonLiveSequences()) {
//				List<String> marking = new LinkedList<>();
//				for (Place p : pair.getFirst()) {
//					for (Place place : d.getPlaces()) {
//						if (p.getLabel().equals(place.getLabel())) {
//							marking.add(place.getLocalID().localNodeID().toString());
//							break;
//						}
//					}
//				}
//				Transition t = pair.getSecond();
//				for (Transition transition : d.getTransitions()) {
//					if (t.getLabel().equals(transition.getLabel())) {
//						information.addNonLiveSequence(marking, transition.getLocalID().localNodeID().toString());
//						break;
//					}
//				}
//			}
//		}
//		if (diagnosis.isBounded() == Boolean.FALSE) {
//			for (Pair<Marking, Transition> pair : diagnosis.getUnboundedSequences()) {
//				List<String> marking = new LinkedList<>();
//				for (Place p : pair.getFirst()) {
//					for (Place place : d.getPlaces()) {
//						if (p.getLabel().equals(place.getLabel())) {
//							marking.add(place.getLocalID().localNodeID().toString());
//						}
//					}
//				}
//				Transition t = pair.getSecond();
//				for (Transition transition : d.getTransitions()) {
//					if (t.getLabel().equals(transition.getLabel())) {
//						information.addUnboundedSequence(marking, transition.getLocalID().localNodeID().toString());
//						break;
//					}
//				}
//			}
//		}
//
		ProcessMiningAlgorithms.postprocessDiagnosisPetrinet(context, (Petrinet) d, labels);

		// Adding extra information
		if (m != null) {
			for (Place p : m) {
				graph.getNode(nodeMap.get(p)).setBoolean(PNConstants.NODE_STARTING_KEY, true);
			}
			markEndingNodes(graph);
		} else {
			markStartingNodes(graph);
			markEndingNodes(graph);
		}

		return graph;
	}

	private static void markEndingNodes(Graph g) {
		Iterator<Node> iter = Utils.getSinkNodes(g);
		while (iter.hasNext()) {
			Node node = iter.next();
			if (NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.PLACE)
				node.setBoolean(PNConstants.NODE_ENDING_KEY, true);
		}
	}

	private static void markStartingNodes(Graph g) {
		Iterator<Node> iter = Utils.getRootNodes(g);
		while (iter.hasNext()) {
			Node node = iter.next();
			if (NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)) == NodeType.PLACE)
				node.setBoolean(PNConstants.NODE_STARTING_KEY, true);
		}
	}
}
