package org.processmining.prabhakardixit.featureSelection;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;

public class CalculateLogHeuristics {
	public static Map<String, Map<String, Double>> followsMap;
	public static Map<String, Map<String, Double>> precedesMap;
	public static Map<String, Map<String, Double>> coOccursMap; //initiate all
	public static Map<String, Double> occursCount = new HashMap<String, Double>();
	public static XLogInfo info;
	public static Set<String> events;
	public static Map<String, Map<String, Double>> directlyFollowsMap;
	public static Map<String, Map<String, Double>> directlyPrecedesMap;
	public static Map<String, Map<String, Double>> eventuallyFollowsMap;
	public static Map<String, Map<String, Double>> eventuallyPrecedesMap;
	public static Map<String, Double> avgOccurrences;
	public static Map<String, Integer> numOfOccurrences;
	public static Map<String, Integer> logSeqInput;

	public static Map<String, Set<String>> attrValMap = new HashMap<String, Set<String>>();

	public static Map<String, Integer> getLogSeqString() {
		return logSeqInput;
	}

	public CalculateLogHeuristics(XLog log) {
		logSeqInput = new HashMap<String, Integer>();
		occursCount = new HashMap<String, Double>();
		followsMap = new HashMap<String, Map<String, Double>>();
		coOccursMap = new HashMap<String, Map<String, Double>>();
		numOfOccurrences = new HashMap<String, Integer>();
		avgOccurrences = new HashMap<String, Double>();
		directlyPrecedesMap = new HashMap<String, Map<String, Double>>();
		directlyFollowsMap = new HashMap<String, Map<String, Double>>();
		eventuallyPrecedesMap = new HashMap<String, Map<String, Double>>();
		eventuallyFollowsMap = new HashMap<String, Map<String, Double>>();
		events = new HashSet<String>();
		Map<String, Integer> logSeq = new HashMap<String, Integer>();
		//Place conformance starts
		Map<String, Map<String, Integer>> preSet = new HashMap<String, Map<String, Integer>>();
		Map<String, Map<String, Integer>> postSet = new HashMap<String, Map<String, Integer>>();
		//Place conformance ends
		initiateInfo(log);
		for (XTrace trace : log) {
			Map<String, Map<String, Double>> followsTraceMap = new HashMap<String, Map<String, Double>>();
			Map<String, Map<String, Double>> precedesTraceMap = new HashMap<String, Map<String, Double>>();
			Map<String, Map<String, Double>> directlyFollowsTraceMap = new HashMap<String, Map<String, Double>>();
			//			Map<String, Map<String, Double>> directlyPrecedesTraceMap = new HashMap<String, Map<String, Double>>();
			Map<String, Map<String, Double>> eventuallyFollowsTraceMap = new HashMap<String, Map<String, Double>>();
			//			Map<String, Map<String, Double>> eventuallyPrecedesTraceMap = new HashMap<String, Map<String, Double>>();
			Map<String, Double> occursInTrace = new HashMap<String, Double>();
			initiate(followsTraceMap, precedesTraceMap);
			initiate(directlyFollowsTraceMap, eventuallyFollowsTraceMap);
			String eventPrev = "";
			String currEvtStr = "";
			String eventSeq = "";
			Set<String> traceEventsTemp = new HashSet<String>();
			setAllAttributes(trace); //update all trace level attr values

			for (XEvent event : trace) {
				currEvtStr = XConceptExtension.instance().extractName(event);
				setAllAttributes(event); //update all event level attr values
				if (traceEventsTemp.add(currEvtStr))
					if (numOfOccurrences.containsKey(currEvtStr))
						numOfOccurrences.put(currEvtStr, numOfOccurrences.get(currEvtStr) + 1);
					else
						numOfOccurrences.put(currEvtStr, 1);
				//make log event seq starts
				if (eventSeq == "")
					eventSeq += currEvtStr;
				else
					eventSeq += ";" + currEvtStr;
				//make log event seq ends

				for (String key : occursInTrace.keySet()) {
					int count = 1;
					if (followsTraceMap.get(key).containsKey(currEvtStr)) {
						count += followsTraceMap.get(key).get(currEvtStr);
					}
					followsTraceMap.get(key).put(currEvtStr, (double) count);
					if (!key.equals(currEvtStr))
						eventuallyFollowsTraceMap.get(key).put(currEvtStr, 1.);
				}
				if (occursInTrace.containsKey(currEvtStr))
					occursInTrace.put(currEvtStr, occursInTrace.get(currEvtStr) + 1);
				else {
					occursInTrace.put(currEvtStr, 1.0);

				}

				if (eventPrev == "") {
					//					if (!preSet.containsKey(currEvtStr))
					//						preSet.put(currEvtStr, new HashMap<String, Integer>());
					//
					//					if (preSet.get(currEvtStr).containsKey("")) //"" indicates start
					//						preSet.get(currEvtStr).put("", preSet.get(currEvtStr).get("") + 1);
					//					else
					//						preSet.get(currEvtStr).put("", 1);
					eventPrev = currEvtStr;
				} else {
					directlyFollowsTraceMap.get(eventPrev).put(currEvtStr, 1.);

					//update Presets
					//					if (!preSet.containsKey(currEvtStr))
					//						preSet.put(currEvtStr, new HashMap<String, Integer>());
					//					if (preSet.get(currEvtStr).containsKey(eventPrev)) //"" indicates start
					//						preSet.get(currEvtStr).put(eventPrev, preSet.get(currEvtStr).get(eventPrev) + 1);
					//					else
					//						preSet.get(currEvtStr).put(eventPrev, 1);
					////					//update postsets
					//					if (!postSet.containsKey(eventPrev))
					//						postSet.put(eventPrev, new HashMap<String, Integer>());
					//					if (postSet.get(eventPrev).containsKey(currEvtStr)) //"" indicates start
					//						postSet.get(eventPrev).put(currEvtStr, postSet.get(eventPrev).get(currEvtStr) + 1);
					//					else
					//						postSet.get(eventPrev).put(currEvtStr, 1);

					eventPrev = currEvtStr;
				}
			}

			if (logSeq.containsKey(eventSeq))
				logSeq.put(eventSeq, logSeq.get(eventSeq) + 1);
			else
				logSeq.put(eventSeq, 1);

			//update the postset with last event
			if (!postSet.containsKey(currEvtStr))
				postSet.put(currEvtStr, new HashMap<String, Integer>());
			if (postSet.get(currEvtStr).containsKey("")) //"" indicates start
				postSet.get(currEvtStr).put("", postSet.get(currEvtStr).get("") + 1);
			else
				postSet.get(currEvtStr).put("", 1);

			//update global follows relations
			for (String key : followsTraceMap.keySet()) {
				for (String traceMappedKey : followsTraceMap.get(key).keySet()) {
					if (key.toString().equals(traceMappedKey))
						continue;
					Double value = 0.;
					if (followsTraceMap.get(key).get(traceMappedKey) > 0) {
						//normalize loops and multiple occurrences
						value = (followsTraceMap.get(key).get(traceMappedKey) - followsTraceMap.get(traceMappedKey)
								.get(key))
								/ (followsTraceMap.get(key).get(traceMappedKey) + followsTraceMap.get(traceMappedKey)
										.get(key));
					}
					value += followsMap.get(key).get(traceMappedKey);
					followsMap.get(key).put(traceMappedKey, value);
				}
			}

			for (String key : occursInTrace.keySet()) {
				occursCount.put(key, occursCount.get(key) + 1);
				if (avgOccurrences.containsKey(key))
					avgOccurrences.put(key, avgOccurrences.get(key) + occursInTrace.get(key)); //add all and divide by # of total loopable occurrences
				else
					avgOccurrences.put(key, 1.);
				for (String coOccursKey : coOccursMap.get(key).keySet())
					if (occursInTrace.containsKey(coOccursKey))
						coOccursMap.get(key).put(coOccursKey, coOccursMap.get(key).get(coOccursKey) + 1);
				//update directly follows
				for (String traceMappedKey : directlyFollowsTraceMap.get(key).keySet()) {
					if (directlyFollowsTraceMap.get(key).get(traceMappedKey) > 0) {
//						System.out.println(key + traceMappedKey);
						if (!key.equals(traceMappedKey))
							directlyFollowsMap.get(key).put(traceMappedKey,
									directlyFollowsMap.get(key).get(traceMappedKey) + 1);
					}
				}
				//update eventually follows
				for (String traceMappedKey : eventuallyFollowsTraceMap.get(key).keySet()) {
					if (eventuallyFollowsTraceMap.get(key).get(traceMappedKey) > 0) {
						eventuallyFollowsMap.get(key).put(traceMappedKey,
								eventuallyFollowsMap.get(key).get(traceMappedKey) + 1);
					}
				}
			}
		}

		for (String key : coOccursMap.keySet()) {
			for (String coOccursKey : coOccursMap.get(key).keySet()) {
				Double updatedFollowsValue = 0.;
				updatedFollowsValue = followsMap.get(key).get(coOccursKey) / coOccursMap.get(key).get(coOccursKey);
				followsMap.get(key).put(coOccursKey, updatedFollowsValue);
				Double value = 0.;
				value = occursCount.get(coOccursKey) + occursCount.get(key) - coOccursMap.get(key).get(coOccursKey);
				value = coOccursMap.get(key).get(coOccursKey) / value;
				coOccursMap.get(key).put(coOccursKey, value);
			}
		}

		for (String key : coOccursMap.keySet()) {
//			System.out.println("key " + key);
			for (String coOccursKey : coOccursMap.get(key).keySet()) {
//				System.out.println("--- CoOccurs " + coOccursKey + " value:  " + coOccursMap.get(key).get(coOccursKey));
//				System.out.println("--- Follows " + coOccursKey + " value: " + followsMap.get(key).get(coOccursKey));
			}
		}

		//make percentages from raw numbers
		for (String occursKey : occursCount.keySet()) {
			for (String mappedKey : directlyFollowsMap.get(occursKey).keySet()) {
				double d = directlyFollowsMap.get(occursKey).get(mappedKey) / occursCount.get(mappedKey); //get avg
				directlyPrecedesMap.get(mappedKey).put(occursKey, d);
				d = directlyFollowsMap.get(occursKey).get(mappedKey) / occursCount.get(occursKey); //get avg
				directlyFollowsMap.get(occursKey).put(mappedKey, d);
			}
			for (String mappedKey : eventuallyFollowsMap.get(occursKey).keySet()) {
				double d = eventuallyFollowsMap.get(occursKey).get(mappedKey) / occursCount.get(mappedKey); //get avg
				eventuallyPrecedesMap.get(mappedKey).put(occursKey, d);
				d = eventuallyFollowsMap.get(occursKey).get(mappedKey) / occursCount.get(occursKey); //get avg
				eventuallyFollowsMap.get(occursKey).put(mappedKey, d);
			}

			double d = avgOccurrences.get(occursKey) / occursCount.get(occursKey); //get avg
			avgOccurrences.put(occursKey, d);
		}

		//update place log seq
		logSeqInput = logSeq;
//		System.out.println("abasdas");

	}

	private void setAllAttributes(XTrace trace) {
		for (String key : trace.getAttributes().keySet()) {
			if (!key.contains("concept")) {
				if (!attrValMap.containsKey(key))
					attrValMap.put(key, new HashSet<String>());
				Set<String> values = new HashSet<String>(attrValMap.get(key));
				values.add(trace.getAttributes().get(key).toString());
				attrValMap.put(key, values);
			}
		}
	}

	private void setAllAttributes(XEvent event) {
		for (String key : event.getAttributes().keySet()) {
			if (!attrValMap.containsKey(key))
				attrValMap.put(key, new HashSet<String>());
			Set<String> values = new HashSet<String>(attrValMap.get(key));
			values.add(event.getAttributes().get(key).toString());
			attrValMap.put(key, values);
		}
	}

	private static void initiate(Map<String, Map<String, Double>> followsMap,
			Map<String, Map<String, Double>> coOccursMap) {
		followsMap.clear();
		coOccursMap.clear();
		for (String evt : events) {
			if (!followsMap.containsKey(evt)) {
				followsMap.put(evt.toString(), new HashMap<String, Double>());
				coOccursMap.put(evt.toString(), new HashMap<String, Double>());
			}
			for (String evt2 : events)
				if (evt != evt2) {
					followsMap.get(evt.toString()).put(evt2, 0.);
					coOccursMap.get(evt.toString()).put(evt2, 0.);
				}
		}

	}

	private static void initiateOccursEvents() {
		for (String event : events)
			occursCount.put(event, 0.);
	}

	private static void initiateInfo(XLog log) {
		final XEventNameClassifier lifecycleClassifier = new XEventNameClassifier();
		info = XLogInfoFactory.createLogInfo(log, lifecycleClassifier);
		//		info.getEventClasses().getClassifier();
		//		
		//		
		//		for(XAttribute x:info.getLog().getGlobalEventAttributes()){
		//		
		//		}

		for (XEventClass event : info.getEventClasses().getClasses()) {
			events.add(event.toString());
			//			events.add(event.toString().replace("+complete", ""));
//			System.out.println(event.getId());
		}
		initiate(followsMap, coOccursMap);
		initiate(directlyFollowsMap, directlyPrecedesMap);
		initiate(eventuallyPrecedesMap, eventuallyFollowsMap);
		initiateOccursEvents();
	}

	public static <K, V extends Comparable<? super V>> Map<K, V> entriesSortedByValues(Map<K, V> map) {

		List<Map.Entry<K, V>> list = new LinkedList<>(map.entrySet());
		Collections.sort(list, new Comparator<Map.Entry<K, V>>() {
			@Override
			public int compare(Map.Entry<K, V> e1, Map.Entry<K, V> e2) {
				return (e2.getValue()).compareTo(e1.getValue());
			}
		});

		Map<K, V> result = new LinkedHashMap<>();
		for (Map.Entry<K, V> entry : list) {
			result.put(entry.getKey(), entry.getValue());
		}

		return result;

	}
	
	public static Map<String, Set<String>> getAttrValuesMap(){
		return attrValMap;
	}
}
